# -*- coding: utf-8 -*-
"""
Manage tables used by Pineboo.

Maintains the definition of a table.

This class maintains the definition of
certain characteristics of a base table
of data.

Additionally it can be used for the definition of
the metadata of a query, see FLTableMetaData :: query ().
"""

from pineboolib.interfaces import itablemetadata
from pineboolib import logging
import copy

from typing import Optional, List, Dict, Union, TYPE_CHECKING
from pineboolib.application.metadata import pnfieldmetadata
from pineboolib.application.metadata import pncompoundkeymetadata

if TYPE_CHECKING:
    from pineboolib.application.metadata import pnrelationmetadata  # noqa # pragma: no cover

LOGGER = logging.get_logger(__name__)


class PNTableMetaData(itablemetadata.ITableMetaData):
    """PNTableMetaData Class."""

    private: "PNTableMetaDataPrivate"

    def __init__(
        self,
        name_or_metadata: Union[str, "PNTableMetaData"] = "",
        alias: Optional[str] = None,
        query_name: Optional[str] = None,
    ) -> None:
        """
        Collect the data to start.

        @param name_or_metadata Name of the table to define
        @param alias Alias ​​of the table, used in forms
        @param query_name (Optional) Name of the query from which you define your metadata
        """

        super().__init__(name_or_metadata, alias, query_name)
        # tmp = None
        if not isinstance(name_or_metadata, str):
            self.inicializeFLTableMetaData(name_or_metadata)

        else:
            if alias is not None and query_name is not None:
                self.inicializeNewFLTableMetaData(name_or_metadata, alias, query_name)
            else:
                self.inicializeFLTableMetaDataP(name_or_metadata)

    def inicializeFLTableMetaData(self, other: Optional["PNTableMetaData"] = None) -> None:
        """
        Initialize the data from another PNTableMetaData.
        """

        self.private = PNTableMetaDataPrivate()
        self.copy(other)

    def inicializeNewFLTableMetaData(
        self, name: str, alias: str, query_name: Optional[str] = None
    ) -> None:
        """
        Initialize the data with the basic information.

        @param name Name of the table to define
        @param alias Alias ​​of the table, used in forms
        @param query_name (Optional) Name of the query from which you define your metadata

        """
        self.private = PNTableMetaDataPrivate(name, alias, query_name)

    def inicializeFLTableMetaDataP(self, name: str) -> None:
        """
        Initialize the private part, without data. Just specify the name.
        """

        self.private = PNTableMetaDataPrivate(name)

        self.private._compound_key = pncompoundkeymetadata.PNCompoundKeyMetaData()

    def name(self) -> str:
        """
        Get the name of the table.

        @return The name of the table described
        """

        return self.private._name

    def setName(self, name: str) -> None:
        """
        Set the name of the table.

        @param name Table name
        """

        # QObject::setName(n);
        self.private._name = name

    def setAlias(self, alias: str) -> None:
        """
        Set the alias.

        @param a Alias
        """

        self.private._alias = alias

    def setQuery(self, query: str) -> None:
        """
        Set the name of the query.

        @param q Query name
        """

        self.private._query = query

    def alias(self) -> str:
        """
        Get the alias associated with the table.

        @return Alias.
        """

        return self.private._alias

    def query(self) -> str:
        """
        Get the name of the query from which it defines its metadata.

        The name will correspond to the definition of a query by
        (.qry file). If the name of the query is defined then
        the name of the table will correspond to the main table of the query
        when referring to several tables.
        """

        return self.private._query

    def isQuery(self) -> bool:
        """
        Get if you define the metadata of a query.
        """

        return True if self.private._query else False

    def addFieldMD(self, field_metadata: "pnfieldmetadata.PNFieldMetaData") -> None:
        """
        Add the description of a field to the list of field descriptions.

        @param f FLFieldMetaData object with the description of the field to add
        """

        # if f is None:
        #     return
        if not field_metadata.metadata():
            field_metadata.setMetadata(self)
        # self.private._field_list.append(field_metadata)
        self.private._fields_dict[field_metadata.name()] = field_metadata
        self.private.addFieldName(field_metadata.name())
        self.private.formatAlias(field_metadata)

        if field_metadata.type() == pnfieldmetadata.PNFieldMetaData.Unlock:
            self.private._field_names_unlock.append(field_metadata.name())
        elif field_metadata.isPrimaryKey():
            self.private._primary_key = field_metadata.name()

    def removeFieldMD(self, field_name: str) -> None:
        """
        Remove the description of a field from the list of field descriptions.

        @param fN Name of the field to be deleted
        """

        if field_name.lower() in self.private._field_names:
            del self.private._fields_dict[field_name.lower()]

        self.private.removeFieldName(field_name)

    def setCompoundKey(
        self, compound_key: Optional["pncompoundkeymetadata.PNCompoundKeyMetaData"]
    ) -> None:
        """
        Set the composite key of this table.

        @param cK FLCompoundKey object with the description of the composite key
        """

        self.private._compound_key = compound_key

    def primaryKey(self, prefix_table=False) -> str:
        """
        Get the name of the field that is the primary key for this table.

        @param prefixTable If TRUE, a prefix with the name of the table is added; tablename.fieldname
        """

        if not self.private._primary_key:
            field_list = self.fieldList()
            length_field_list = len(field_list)
            for field in field_list:
                if field.type() == "serial" or length_field_list == 1:
                    LOGGER.debug(
                        "Forzando %s(serial) como primaryKey de %s", field.name(), self.name()
                    )
                    self.private._primary_key = field.name()
                    field.setIsPrimaryKey(True)
                    break

            if not self.private._primary_key:
                raise Exception("No primaryKey in %s" % self.private._name)

        return (
            "%s.%s" % (self.private._name, self.private._primary_key)
            if prefix_table and "." not in self.private._primary_key
            else self.private._primary_key
        )

    def fieldNameToAlias(self, field_name: str = "") -> str:
        """
        Get the alias of a field from its name.

        @param fN Field name
        """

        return (
            self.private._field_alias_map[field_name]
            if field_name in self.private._field_alias_map
            else ""
        )

    def fieldAliasToName(self, alias_name: str = "") -> str:
        """
        Get the name of a field from its alias.

        @param aN Field alias name
        """

        return (
            self.private._alias_field_map[alias_name]
            if alias_name in self.private._alias_field_map.keys()
            else ""
        )

    def fieldType(self, field_name: str = "") -> int:
        """
        Get the type of a field from its name.

        @param fN Field name
        """

        ret_ = 0
        type_ = (
            self.private._fields_dict[field_name].type()
            if field_name in self.private._field_names
            else None
        )

        if type_:
            if type_ in ("string", "counter", "timestamp"):
                ret_ = 3
            elif type_ == "stringlist":
                ret_ = 4
            elif type_ == "pixmap":
                ret_ = 6
            elif type_ == "uint":
                ret_ = 17
            elif type_ == "bool":
                ret_ = 18
            elif type_ == "double":
                ret_ = 19
            elif type_ == "date":
                ret_ = 26
            elif type_ == "time":
                ret_ = 27
            elif type_ == "serial":
                ret_ = 100
            elif type_ == "unlock":
                ret_ = 200
            elif type_ == "check":
                ret_ = 300
            else:
                # FIXME: Falta int
                LOGGER.warning("FIXME:: No hay definido un valor numérico para el tipo %s", type_)

        return ret_

    def fieldIsPrimaryKey(self, field_name: str = "") -> bool:
        """
        Get if a field is the primary key from its name.

        @param fN Field name.
        """

        return field_name == self.private._primary_key

    def fieldIsIndex(self, field_name: str = "") -> int:
        """
        Get if a field is index based on its name.

        @param fN Field name.
        """

        return self.fieldNames().index(field_name) if field_name in self.fieldNames() else -1

    def fieldIsCounter(self, field_name: str = "") -> bool:
        """
        Get if a field is a counter.

        @param fN Field name.
        @author Andrés Otón Urbano (baxas@eresmas.com)
        """

        return (
            self.private._fields_dict[field_name].isCounter()
            if field_name in self.private._field_names
            else False
        )

    def fieldAllowNull(self, field_name: str = "") -> bool:
        """
        Get if a field can be null.

        @param fN Field name
        """

        return (
            self.private._fields_dict[field_name].allowNull()
            if field_name in self.private._field_names
            else False
        )

    def fieldIsUnique(self, field_name: str = "") -> bool:
        """
        Get if a field is unique from its name.

        @param fN Field name.
        """

        return (
            self.private._fields_dict[field_name].isUnique()
            if field_name in self.private._field_names
            else False
        )

    def fieldTableM1(self, field_name: str = "") -> str:
        """
        Get the name of the foreign table related to a field in this table by an M1 relationship (many to one).

        @param fN Field of the relation M1 of this table, which is supposed to be related
            with another field from another table.
        @return The name of the related table M1, if there is a relationship for the field, or a string
            empty without the field is not related.
        """

        if field_name in self.private._field_names:
            relation_ = self.private._fields_dict[field_name].relationM1()
            if relation_:
                return relation_.foreignTable()

        return ""

    def fieldForeignFieldM1(self, field_name: str = "") -> str:
        """
        Get the name of the foreign table field related to the one indicated by an M1 relationship (many still).

        @param fN Field of the relation M1 of this table, which is supposed to be related
            with another field from another table.
        @return The name of the foreign field related to the indicated.
        """
        if field_name in self.private._field_names:
            relation_ = self.private._fields_dict[field_name].relationM1()
            if relation_:
                return relation_.foreignField()

        return ""

    def relation(
        self, field_name: str, foreign_field: str, foreign_table: str
    ) -> Optional["pnrelationmetadata.PNRelationMetaData"]:
        """
        Get the relationship object that defines two fields.

        @param field_name Field name of this table that is part of the relationship.
        @param foreign_field Name of the foreign field to this table that is part of the relationship.
        @param foreign_table Name of the foreign table.
        @return Returns a FLRelationMetaData object with the relationship information, provided
            when it exists If it does not exist, it returns False.
        """

        if field_name in self.private._field_names:
            relation_m1 = self.private._fields_dict[field_name].relationM1()
            if relation_m1:
                if (
                    relation_m1.foreignField() == foreign_field
                    and relation_m1.foreignTable() == foreign_table
                ):
                    return relation_m1

            else:
                for relation_1m in list(self.private._fields_dict[field_name].relationList()):
                    if (
                        relation_1m.foreignField() == foreign_field
                        and relation_1m.foreignTable() == foreign_table
                    ):
                        return relation_1m

        return None

    def fieldLength(self, field_name: str = "") -> int:
        """
        Get the length of a field from its name.

        @param fN Field name.
        @return field length.
        """

        return (
            self.private._fields_dict[field_name].length()
            if field_name in self.private._field_names
            else 0
        )

    def fieldPartInteger(self, field_name: str = "") -> int:
        """
        Get the number of digits of the entire part of a field from its name.

        @param fN Field name.
        @return integer length.
        """

        return (
            self.private._fields_dict[field_name].partInteger()
            if field_name in self.private._field_names
            else 0
        )

    def fieldPartDecimal(self, field_name: str = "") -> int:
        """
        Get the number of digits of the decimal part of a field from its name.

        @param fN Field name.
        @return part decimal length.
        """

        return (
            self.private._fields_dict[field_name].partDecimal()
            if field_name in self.private._field_names
            else 0
        )

    def fieldCalculated(self, field_name: str = "") -> bool:
        """
        Get if a field is calculated.

        @param fN Field name.
        """

        return (
            self.private._fields_dict[field_name].calculated()
            if field_name in self.private._field_names
            else False
        )

    def fieldVisible(self, field_name: str = "") -> bool:
        """
        Get if a field is visible.

        @param fN Field name.
        """

        return (
            self.private._fields_dict[field_name].visible()
            if field_name in self.private._field_names
            else False
        )

    def field(self, field_name: str = "") -> Optional["pnfieldmetadata.PNFieldMetaData"]:
        """
        Get the metadata of a field.

        @param fN Field name.
        @return A FLFieldMetaData object with the information or metadata of a given field.
        """

        return (
            self.private._fields_dict[field_name]
            if field_name in self.private._field_names
            else None
        )

    def fieldList(self, only_names: bool = False) -> List["pnfieldmetadata.PNFieldMetaData"]:
        """
        Return a list of field definitions.

        @return Object with the table field deficits list
        """

        return self.fieldNames() if only_names else list(self.private._fields_dict.values())  # type: ignore [return-value, arg-type]

    def fieldListArray(self, prefix_table: bool = False) -> List[str]:
        """
        To get a string with the names of the fields separated by commas.

        @param prefix_table If TRUE a prefix is ​​added to each field with the name of the table; tablename.fieldname
        @return String with the names of the fields separated by commas.
        """

        return [
            "%s.%s" % (self.private._name, field_name) if prefix_table else field_name
            for field_name in self.private._field_names
        ]

    # def fieldListObject(self):
    #    #print("FiledList count", len(self.private._field_list))
    #    return self.private._field_list

    def indexPos(self, field_name: str) -> int:
        """
        Return the position of a field in the real order.

        @param field_name. Field Name.
        @return position index or None.
        """

        return self.fieldIsIndex(field_name)

    def fieldListOfCompoundKey(
        self, field_name: str = ""
    ) -> Optional[List["pnfieldmetadata.PNFieldMetaData"]]:
        """
        Get the list of fields of a compound key, from the name of a field that you want to find out if it is in that compound key.

        @param fN Name of the field you want to find out if it belongs to a compound key.
        @return If the field belongs to a composite key, it returns the list of fields
          that form said composite key, including the consulted field. If
          that the consulted field does not belong to any compound key returns None
        """

        if self.private._compound_key:
            if self.private._compound_key.hasField(field_name):
                return self.private._compound_key.fieldList()
        return None

    def fieldNames(self) -> List[str]:
        """
        Get a list of texts containing the names of fields separated by commas.

        The order of the fields from left to right corresponds to the order in which
        have been added with the addFieldMD () or addFieldName () method

        @return field name list.
        """

        return list(self.private._field_names)

    def fieldNamesUnlock(self) -> List[str]:
        """
        List of field names in the table that are of type PNFieldMetaData :: Unlock.

        @return field name list.
        """

        return self.private._field_names_unlock

    def concurWarn(self) -> bool:
        """
        Return concurWarn flag.

        @return True or False
        """

        return self.private._concur_warn

    def setConcurWarn(self, state: bool = True) -> None:
        """
        Enable concurWarn flag.

        @param state. True or False.
        """

        self.private._concur_warn = state

    def detectLocks(self) -> bool:
        """
        Return lock detection flag.

        @return b. True or False.
        """

        return self.private._detect_locks

    def setDetectLocks(self, state: bool = True) -> None:
        """
        Enable lock detection flag.

        @return b. True or False.
        """

        self.private._detect_locks = state

    def FTSFunction(self) -> str:
        """
        Return function name to call for Full Text Search.

        @return function name or None.
        """

        return self.private.full_text_search_function

    def setFTSFunction(self, full_text_search_function: str) -> None:
        """
        Set the function name to call for Full Text Search.

        @param ftsfun. function name.
        """

        self.private.full_text_search_function = full_text_search_function

    def inCache(self) -> bool:
        """
        Return if the metadata is cached (FLManager :: cacheMetaData_).

        @return True or False.
        """

        return self.private._in_cache if self.private else False

    def setInCache(self, state: bool = True) -> None:
        """
        Set the metadata is cached (FLManager :: cacheMetaData_).

        @return True or False.
        """

        self.private._in_cache = state

    def copy(self, other: Optional["PNTableMetaData"] = None) -> None:
        """
        Copy the values of a PNFieldMetaData from another.

        @param other. PNTableMetaData.
        """

        if other is None or other == self:
            return

        self.private = copy.copy(other.private)

    def indexFieldObject(self, position: int) -> "pnfieldmetadata.PNFieldMetaData":
        """
        Return the PNFieldMetaData of the given field.

        @param i. Position.
        @return PNfieldMetadata.
        """
        if position < 0 or position >= len(self.private._field_names):
            raise ValueError("Value n:%s out of bounds" % position)

        return self.private._fields_dict[self.private._field_names[position]]

    def __bool__(self) -> bool:
        """Return if exists."""

        return True

    def setCachedFields(self, cached_fields: str) -> None:
        """Set cached fields."""
        self.private._cached_fields = cached_fields

    def useCachedFields(self) -> bool:
        """Return if use cached fields."""

        return self.private._cached_fields != ""

    def cachedFields(self) -> str:
        """Return cached fields."""

        return self.private._cached_fields


class PNTableMetaDataPrivate:
    """PNTableMetaData Class."""

    """
    Nombre de la tabla
    """

    _name: str

    """
    Alias de la tabla
    """
    _alias: str

    """
    Diccionario de campos que tiene esta tabla
    """
    # _field_list: List["pnfieldmetadata.PNFieldMetaData"]

    _fields_dict: Dict[str, "pnfieldmetadata.PNFieldMetaData"]

    """
    Clave compuesta que tiene esta tabla
    """
    _compound_key: Optional["pncompoundkeymetadata.PNCompoundKeyMetaData"] = None

    """
    Nombre de la consulta (fichero .qry) de la que define los metadatos
    """
    _query: str

    """
    Cadena de texto con los nombre de los campos separados por comas
    """
    _field_names: List[str] = []

    """
    Mapas alias<->nombre
    """
    _alias_field_map: Dict[str, str]
    _field_alias_map: Dict[str, str]

    """
    Lista de nombres de campos de la tabla que son del tipo FLFieldMetaData::Unlock
    """
    _field_names_unlock: List[str] = []

    """
    Clave primaria
    """
    _primary_key: Optional[str]

    """
    Indica si se debe avisar de colisión de concurrencia entre sesiones.

    Si este flag es true y dos o mas sesiones/usuarios están modificando los
    mismos campos,al validar un formulario (FLFormRecordDB::validateForm)
    mostrará un aviso de advertencia.

    Ver también FLSqlCursor::concurrencyFields().
    """
    _concur_warn: bool

    """
    Indica si se deben comprobar riesgos de bloqueos para esta tabla

    Si este flag es true FLSqlCursor::commitBuffer() chequeará siempre
    los riesgos de bloqueo para esta tabla.

    Ver también FLSqlDatabase::detectRisksLocks
    """
    _detect_locks: bool

    """
    Indica el nombre de función a llamar para la búsqueda con Full Text Search
    """
    full_text_search_function: str

    """
    Indica si lo metadatos están en caché (FLManager::cacheMetaData_)
    """
    _in_cache: bool

    count_ = 0

    _cached_fields: str

    def __init__(
        self,
        name: Optional[str] = None,
        alias: Optional[str] = None,
        qry_name: Optional[str] = None,
    ) -> None:
        """
        Initialize the class.

        @param nane metadata name.
        @param alias metadata alias.
        @param qry_name query string.
        """
        self._name = ""
        self._primary_key = None
        self._concur_warn = False
        self.full_text_search_function = ""
        # self._field_list = []
        self._fields_dict = {}
        self._field_names = []
        self._field_names_unlock = []
        self._alias_field_map = {}
        self._field_alias_map = {}
        self._detect_locks = True
        self._query = ""
        self._in_cache = False
        self._cached_fields = ""
        # print("Vaciando field list ahora",  len(self._field_list))

        if name:
            if alias:
                self.inicializeNewFLTableMetaDataPrivate(name, alias, qry_name)
            else:
                self.inicializeFLTableMetaDataPrivateS(name)
        else:
            self.inicializeFLTableMetaDataPrivate()

        self.count_ += 1

    def inicializeFLTableMetaDataPrivate(self) -> None:
        """
        Initialize class ends with empty data.
        """

        self._compound_key = None

    def inicializeNewFLTableMetaDataPrivate(
        self, name: str, alias: str, query: Optional[str] = None
    ) -> None:
        """
        Initialize the class end with data.

        @param name metadata name.
        @param alias metadata alias.
        @param query query string.
        """

        self._name = name.lower()
        self._alias = alias
        self._compound_key = None
        if query is not None:
            self._query = query
        self._concur_warn = False
        self._detect_locks = False

    def inicializeFLTableMetaDataPrivateS(self, name: str) -> None:
        """
        Initialize the class end with basic data.

        @param name metadata name.
        """

        self._name = str(name)
        self._alias = self._name

    def addFieldName(self, name: str) -> None:
        """
        Add the name of a field to the field name string, see fieldNames().

        @param name Field Name.
        """

        self._field_names.append(name.lower())

    def removeFieldName(self, name: str) -> None:
        """
        Remove the name of a field from the field name string, see fieldNames().

        @param name Field Name
        """

        if name in self._field_names:
            self._field_names.remove(name)

        if name in self._field_names_unlock:
            self._field_names_unlock.remove(name)
        if self._primary_key == name:
            self._primary_key = None

    def formatAlias(self, field_object: Optional["pnfieldmetadata.PNFieldMetaData"] = None) -> None:
        """
        Format the alias of the indicated field to avoid duplicates.

        @param f Object field whose alias you want to format
        """

        if field_object is not None:
            alias = field_object.alias()
            field_name = field_object.name()

            if alias in self._alias_field_map:
                alias = "%s(%s)" % (alias, str(len(self._alias_field_map) + 1))

            field_object.private.alias_ = alias

            self._alias_field_map[alias] = field_name
            self._field_alias_map[field_name] = alias

    def clearFieldList(self) -> None:
        """
        Clear the list of field definitions.
        """

        # self._field_list = []
        self._fields_dict = {}
        self._field_names = []



"""
Module for PNSignals class.
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.interfaces import isqlcursor


class PNSignals(object):
    """
    Signals for database.

    Used in flapplication.
    Filters out transaction signals.
    """

    notify_begin_transaction_: bool
    notify_end_transaction_: bool
    notify_roll_back_transaction_: bool

    def __init__(self):
        """
        Create new PNSignals.

        It is treated like a singleton. See pineboolib.database.db_signals
        """
        super().__init__()

        self.notify_begin_transaction_ = False
        self.notify_end_transaction_ = False
        self.notify_roll_back_transaction_ = False

    def emitTransactionBegin(self, cursor: "isqlcursor.ISqlCursor") -> None:
        """Emit transaction begin signal."""
        if self.notify_begin_transaction_:
            cursor.transactionBegin.emit()

    def emitTransactionEnd(self, cursor: "isqlcursor.ISqlCursor") -> None:
        """Emit transaction end signal."""
        if self.notify_end_transaction_:
            cursor.transactionEnd.emit()

    def emitTransactionRollback(self, cursor: "isqlcursor.ISqlCursor") -> None:
        """Emit transaction begin rollback."""
        if self.notify_roll_back_transaction_:
            cursor.transactionRollback.emit()



"""Test ThreadSession module."""

import unittest
import threading
import time

from pineboolib.loader.main import init_testing, finish_testing
from pineboolib import application
from pineboolib.qsa import qsa

SESSION_LIST = []


class TestThreadSession(unittest.TestCase):
    """TestQueryOrm Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()
        application.PROJECT.conn_manager.manager().createTable("fltest4")
        application.PROJECT.conn_manager.manager().createTable("fltest5")

    def test_basic_1(self) -> None:
        """Test basic 1."""

        self.assertTrue(prueba(1))

    def test_basic_2(self) -> None:
        """Test basic 2."""

        self.assertTrue(prueba2())

    def test_basic_3(self) -> None:
        """Test basic 3."""
        self.assertFalse(prueba3())

    def test_basic_4(self) -> None:
        """Test basic 4."""

        for num in range(50):
            thr = threading.Thread(target=massive, args=(num,))
            thr.start()

        while len(SESSION_LIST) < 50:
            time.sleep(0.1)

        for session in SESSION_LIST:
            self.assertFalse(session.transaction)  # type: ignore [union-attr] # noqa: F821

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()


@qsa.atomic()  # type: ignore [misc] # noqa: F821
def massive(value: int):
    """Massive test."""
    print(
        "____inicio",
        value,
        qsa.session_atomic(),  # type: ignore [union-attr] # noqa: F821
        qsa.session_atomic().transaction,  # type: ignore [union-attr] # noqa: F821
    )
    if not qsa.session_atomic().transaction:  # type: ignore [union-attr] # noqa: F821
        raise Exception("Transaction is empty!")

    SESSION_LIST.append(qsa.session_atomic())


@qsa.atomic()  # type: ignore [misc] # noqa: F821
def prueba(value: int):
    """Prueba function."""

    mng_ = application.PROJECT.conn_manager

    return mng_.useConn("default").session() is qsa.session_atomic()


@qsa.atomic("dbaux")  # type: ignore [misc] # noqa: F821
def prueba2():
    """Prueba2 function."""
    mng_ = application.PROJECT.conn_manager

    return mng_.useConn("dbaux").session() is qsa.session_atomic("dbaux")


@qsa.atomic("dbaux")  # type: ignore [misc] # noqa: F821
def prueba3():
    """Prueba3 function."""
    obj_ = qsa.orm.fltest4()
    return obj_.session == qsa.session_atomic("dbaux")



"""
Interfaces module.

This module contains some common class interfaces, so they can be interchanged.
"""

from pineboolib.interfaces.ifieldmetadata import IFieldMetaData  # noqa: F401
from pineboolib.interfaces.itablemetadata import ITableMetaData  # noqa: F401
from pineboolib.interfaces.imanager import IManager  # noqa: F401



"""Test query_orm module."""

import unittest

from pineboolib.loader.main import init_testing, finish_testing
from pineboolib import application
from pineboolib.qsa import qsa


class TestQueryOrm(unittest.TestCase):
    """TestQueryOrm Class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()
        application.PROJECT.conn_manager.manager().createTable("fltest4")
        application.PROJECT.conn_manager.manager().createTable("fltest5")

    def test_session(self) -> None:
        """Test session query."""

        session = qsa.session()
        session.begin()
        class_area = qsa.orm_("fltest4")
        obj_area = class_area()
        session.close()
        obj_area.idarea = "E"
        obj_area.other_field = "DSGDSGSDG**"
        self.assertTrue(obj_area.save())

        class_area = qsa.orm_("fltest4")
        result = class_area.query().filter(class_area.idarea == "E").first()
        self.assertTrue(result)

    def test_delete(self) -> None:
        """Test delete with children."""

        session = qsa.thread_session_new()
        session.begin()
        class_area = qsa.orm_("fltest4")
        obj_area = class_area()

        obj_area.idarea = "E"
        obj_area.other_field = "DSGDSGSDG**"
        self.assertTrue(obj_area.save())

        class_child = qsa.orm_("fltest5")
        child_1 = class_child()
        child_2 = class_child()

        child_1.idmodulo = "A"
        child_2.idmodulo = "B"
        child_1.idarea = "E"
        child_2.idarea = "E"
        self.assertTrue(child_1.save())
        child_1.session.commit()
        session.begin()
        self.assertTrue(child_2.save())
        child_2.session.commit()

        self.assertEqual(len(class_child.query().all()), 2)

        # session.commit()
        session.begin()
        self.assertTrue(class_area.query().filter(class_area.idarea == "E").first())
        # session = qsa.session()
        # new_obj = class_area.get("E")
        self.assertEqual(len(class_child.query().all()), 2)
        lista = obj_area.query().filter(class_area.idarea == "E")
        for obj in lista:
            obj.delete()

        session.commit()
        # session.commit()
        self.assertEqual(len(class_child.query().all()), 0)

        self.assertFalse(class_area.get("E"))
        session.close()

    def test_delete2(self) -> None:
        """Test delete with children."""

        session = qsa.thread_session_new()
        session.begin()
        class_area = qsa.orm_("fltest4")
        obj_area = class_area()

        obj_area.idarea = "E"
        obj_area.other_field = "DSGDSGSDG**"
        self.assertTrue(obj_area.save())

        class_child = qsa.orm_("fltest5")
        child_1 = class_child()
        child_2 = class_child()

        child_1.idmodulo = "A"
        child_2.idmodulo = "B"
        child_1.idarea = "E"
        child_2.idarea = "E"
        self.assertTrue(child_1.save())
        child_1.session.commit()
        session.begin()
        self.assertTrue(child_2.save())
        child_2.session.commit()

        self.assertEqual(len(class_child.query().all()), 2)

        # session.commit()
        session.begin()
        self.assertTrue(class_area.query().filter(class_area.idarea == "E").first())
        # session = qsa.session()
        # new_obj = class_area.get("E")
        self.assertEqual(len(class_child.query().all()), 2)
        lista = class_child.query().filter(class_child.idarea == "E").all()
        for obj in lista:
            self.assertTrue(class_area.query().filter(class_area.idarea == obj.idarea).first())
            obj.delete()

        session.commit()
        # session.commit()
        self.assertEqual(len(class_child.query().all()), 0)

        self.assertFalse(class_area.get("E"))
        session.close()

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""

        finish_testing()



# -*- coding: utf-8 -*-
"""PNAction Module."""

from pineboolib.core.utils.struct import ActionStruct
from pineboolib import logging
from typing import Union, Any

LOGGER = logging.get_logger(__name__)


class PNAction(object):
    """
    PNAction Class.

    This class contains information on actions to open forms.

    It is used to automatically link forms with your script,
    interface and source table.

    @author InfoSiAL S.L.
    """

    """
    Nombre de la accion
    """
    _name: str

    """
    Nombre del script asociado al formulario de edición de registros
    """
    _script_form_record: str

    """
    Nombre del script asociado al formulario maestro
    """
    _script_form: str

    """
    Nombre de la tabla origen para el formulario maestro
    """
    _table: str

    """
    Nombre del formulario maestro
    """
    _form: str

    """
    Nombre del formulario de edición de registros
    """
    _form_record: str

    """
    Texto para la barra de título del formulario maestro
    """
    _caption: str

    """
    Descripción
    """
    _description: str

    """
    Class
    """
    _class: str

    def __init__(self, action: Union[str, ActionStruct]) -> None:
        """Initialize."""

        self._name = ""
        self._caption = ""
        self._description = ""
        self._form = ""
        self._form_record = ""
        self._script_form = ""
        self._script_form_record = ""
        self._table = ""
        self._class = ""

        if isinstance(action, str):
            self.setName(action)

        elif isinstance(action, ActionStruct):
            self.setName(action._name)
            self.setScriptForm(action._master_script)
            self.setScriptFormRecord(action._record_script)
            self.setForm(action._master_form)
            self.setFormRecord(action._record_form)
            self.setCaption(action._alias)
            self.setTable(action._table)
        else:
            raise Exception("Unsupported action %r" % action)

    def __repr__(self):
        """Return the values ​​in a text string."""

        return self._name

    def __str__(self):
        """Return the values ​​in a text string."""

        return self._name

    def __eq__(self, other: Any) -> bool:
        """Return compare result."""

        if isinstance(other, str):
            return self._name == other
        else:
            return self is other

    def setName(self, name: str) -> None:
        """Set the name of the action."""

        self._name = name

    def setScriptFormRecord(self, script_form_record: str) -> None:
        """Set the name of the script associated with the record editing form."""

        self._script_form_record = "%s.qs" % script_form_record if script_form_record else ""

    def setScriptForm(self, script_form: str) -> None:
        """Set the name of the script associated with the master form."""

        self._script_form = "%s.qs" % script_form if script_form else ""

    def setTable(self, table: str) -> None:
        """Set the name of the source table of the master form."""

        self._table = table

    def setClass_(self, class_: str) -> None:
        """Set the name of class script."""

        self._class = class_ if class_ else ""

    def setForm(self, form: str) -> None:
        """Set the name of the master form."""

        self._form = "%s.ui" % form if form else ""

    def setFormRecord(self, form_record: str) -> None:
        """Set the name of the record editing form."""

        self._form_record = "%s.ui" % form_record if form_record else ""

    def setCaption(self, caption: str) -> None:
        """Set the text of the title bar of the master form."""

        self._caption = caption

    def setDescription(self, description: str) -> None:
        """Set description."""

        self._description = description

    def name(self) -> str:
        """Get the name of the action."""

        return self._name

    def scriptFormRecord(self) -> str:
        """Get the name of the script associated with the record editing form."""

        return self._script_form_record

    def scriptForm(self) -> str:
        """Get the name of the script associated with the master form."""

        return self._script_form

    def table(self) -> str:
        """Get the table associated with the action."""

        return self._table

    def caption(self) -> str:
        """Get the text from the form's title bar."""

        return self._caption

    def description(self) -> str:
        """Get the description."""

        return self._description

    def form(self) -> str:
        """Get the name of the mestro form."""

        return self._form

    def formRecord(self) -> str:
        """Get the name of the record editing form."""

        return self._form_record

    def class_(self) -> str:
        """Get class script."""

        return self._class



"""Preload Module."""

from pineboolib import logging
from typing import Container, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from pineboolib.application import projectmodule  # pragma: no cover

LOGGER = logging.get_logger(__name__)


def preload_actions(
    project: "projectmodule.Project", forceload: Optional[Container] = None
) -> None:
    """
    Preload actions for warming up the pythonizer cache.

    forceload: When passed an string, it filters and loads all
        actions that match "*forceload*". If None, all actions
        are loaded.
    """
    LOGGER.info("Precarga ...")
    for action in project.actions:
        if forceload and action not in forceload:
            continue
        LOGGER.debug("* * * Cargando acción %s . . . " % action)
        try:
            project.actions[action].load_master_form()
        except Exception:
            LOGGER.exception("Failure trying to load action %s", action)
            project.conn_manager.mainConn().rollback()  # FIXME: Proper transaction handling using with context
        try:
            project.actions[action].load_record_form()
        except Exception:
            LOGGER.exception("Failure trying to loadRecord action %s", action)
            project.conn_manager.mainConn().rollback()  # FIXME: Proper transaction handling using with context



# -*- coding: utf-8 -*-
"""
AQS package.

Main entrance to the different AQS resources.
"""

from PyQt6 import QtCore, QtWidgets, QtGui, QtXml  # type: ignore[import]

from pineboolib.core.utils import logging

from typing import Any, Optional, Union, cast

from pineboolib.fllegacy.aqsobjects import aqshttp, aqods

LOGGER = logging.get_logger(__name__)


class SMTP(object):
    """Smtp enumerate class."""

    SmtpSslConnection: int = 1
    SmtpTlsConnection: int = 2
    SmtpAuthPlain: int = 1
    SmtpAuthLogin: int = 2
    SmtpSendOk: int = 11
    SmtpError: int = 7
    SmtpMxDnsError: int = 10
    SmtpSocketError: int = 12
    SmtpAttachError: int = 15
    SmtpServerError: int = 16
    SmtpClientError: int = 17


class Docker(object):
    """Docker enumerate class."""

    LeftDockWidgetArea = QtCore.Qt.DockWidgetArea.LeftDockWidgetArea
    InDock: str = "InDock"
    OutSideDock: str = "OutSideDock"


class FLTableDB(object):
    """Fltabledb enumerate class."""

    RefreshData = [False, True]


class PrinterColorMode(object):
    """PrintColorMode enumerate class."""

    PrintGrayScale = 0
    PrintColor = 1


class Events(object):
    """AQS Events."""

    Close = QtGui.QCloseEvent
    Show = QtGui.QShowEvent
    ContextMenu = QtGui.QContextMenuEvent


class AQSClass(SMTP, Docker, FLTableDB, PrinterColorMode, aqods.OdsStyleFlags, Events):
    """AQS Class."""

    Box = None
    Plain = None
    StFailed = None

    def __getattr__(self, name: str) -> Any:
        """
        Return the attributes of the main classes, if it is not in the class.

        @param name. Attribute name.
        @return Attribute or None.
        """

        if name == "DockLeft":
            name = "LeftDockWidgetArea"
            return getattr(self, name)
        elif name == "WordBreak":
            name = "TextWordWrap"
            return getattr(self, name)

        ret_ = getattr(QtCore.Qt, "%sOrder" % name, None)

        if ret_ is None:
            ret_ = getattr(QtGui, "Q%sEvent" % name, None)

        if ret_ is None:
            for lib in [
                QtWidgets.QFrame,
                QtWidgets.QLabel,
                QtWidgets.QSizePolicy,
                QtCore.Qt.TextFlag,
                QtCore.Qt,
                QtCore.Qt.TextFormat,
                QtCore.Qt.CursorShape,
                QtCore.QEvent,
                QtCore.QEvent.Type,
            ]:
                ret_ = getattr(lib, name, None)
                if ret_ is not None:
                    break

        if ret_ is None:
            ret_ = getattr(aqshttp.AQSHttp(), "Http%s" % name, None)

        if ret_ is not None:
            LOGGER.warning("AQS: Looking up attr: %r -> %r  (Please set these in AQS)", name, ret_)
            return ret_

        LOGGER.warning("AQS: No se encuentra el atributo %s", name)

    @staticmethod
    def ColorDialog_getColor(
        color: Optional[Union[int, str, QtGui.QColor]] = None,
        parent: Optional[QtWidgets.QWidget] = None,
        name: Optional[str] = None,
    ) -> Any:
        """
        Display the color selection dialog.

        @param color. Specify the initial color.
        @param parent. Parent.
        @param name. deprecated. Parameter used for compatibility.
        """

        if color is None:
            qcolor = QtGui.QColor("black")
        elif not isinstance(color, QtGui.QColor):
            qcolor = QtGui.QColor(color)
        else:
            qcolor = color

        color_dialog = QtWidgets.QColorDialog(qcolor, parent)
        return color_dialog.getColor()

    @classmethod
    def toXml(
        cls,
        obj_: QtCore.QObject,
        include_children: bool = True,
        include_complex_types: bool = False,
    ) -> QtXml.QDomDocument:
        """
        Convert an object to xml.

        @param obj_. Object to be processed
        @param include_children. Include children of the given object
        @param include_complex_types. Include complex children
        @return xml of the given object
        """

        xml_ = QtXml.QDomDocument()

        if not obj_:
            return xml_

        element = xml_.createElement(type(obj_).__name__)
        element.setAttribute("class", type(obj_).__name__)
        xml_.appendChild(element)

        _meta = obj_.metaObject()

        i = 0
        # _p_properties = []
        for i in range(_meta.propertyCount()):  # type: ignore [union-attr]
            meta_prop = _meta.property(i)  # type: ignore [union-attr]
            # if meta_prop.name() in _p_properties:
            #    i += 1
            #    continue

            # _p_properties.append(meta_prop.name())

            val = getattr(obj_, meta_prop.name(), None)  # type: ignore [arg-type]
            try:
                if val is not None:
                    val = val()
            except Exception:
                pass

            if val is None:
                i += 1
                continue

            val = str(val)

            if not val and not include_complex_types:
                i += 1
                continue
            element.setAttribute(meta_prop.name(), val)

            i += 1

        if include_children:
            for child in obj_.children():
                itd = cls.toXml(child, include_children, include_complex_types)
                xml_.firstChild().appendChild(itd.firstChild())
        return xml_

    @staticmethod
    def pixmap_fromMimeSource(name: str) -> QtGui.QPixmap:
        """
        Get a QtGui.QPixmap of a given file name.

        @param name. File Name
        @return QtGui.QPixmap
        """

        from pineboolib.core.utils import utils_base

        return utils_base.pixmap_from_mime_source(name)

    Pixmap_fromMineSource = pixmap_fromMimeSource

    @classmethod
    def sha1(cls, byte_array: bytes) -> str:
        """
        Return the sha1 of a set of bytes.

        @param byte_array: bytes to process.
        @return sha1 string
        """

        qbyte = QtCore.QByteArray(cast(QtCore.QByteArray, byte_array))
        hash = QtCore.QCryptographicHash(QtCore.QCryptographicHash.Algorithm.Sha1)
        hash.addData(qbyte.data())
        return hash.result().toHex().data().decode("utf-8").upper()

    @classmethod
    def Application_setOverrideCursor(cls, shape: "QtGui.QCursor", replace: bool = False) -> None:
        """
        Set override cursor.

        @param. shape. QtGui.QCursor instance to override.
        @param. replace. Not used.
        """

        QtWidgets.QApplication.setOverrideCursor(shape)

    @classmethod
    def Application_restoreOverrideCursor(cls) -> None:
        """Restore override cursor."""
        QtWidgets.QApplication.restoreOverrideCursor()

    @classmethod
    def TextCodec_codecForName(cls, codec_name: str) -> str:
        """Return codec name."""

        return codec_name


AQS = AQSClass()



"""Flmanager module."""

# -*- coding: utf-8 -*-

from PyQt6 import QtCore, QtXml  # type: ignore[import]

from pineboolib.core import decorators
from pineboolib.core.utils import utils_base

from pineboolib.application.metadata import (
    pncompoundkeymetadata,
    pntablemetadata,
    pnrelationmetadata,
    pnfieldmetadata,
)

from pineboolib.application.database import pnsqlquery, pngroupbyquery, pnsqlcursor
from pineboolib.application.utils import xpm, convert_flaction
from pineboolib.application import qsadictmodules

from pineboolib.interfaces import IManager

from pineboolib import logging, application
from pineboolib.fllegacy import flutil

import copy
import os

from xml.etree import ElementTree

from typing import Optional, Union, Any, Dict, List, TYPE_CHECKING

if TYPE_CHECKING:
    from pineboolib.interfaces import iconnection  # pragma: no cover
    from pineboolib.application.metadata import pnaction  # pragma: no cover
    from pineboolib.application.acls import pnaccesscontrollists  # pragma: no cover


LOGGER = logging.get_logger(__name__)

# FIXME: This class is emulating Eneboo, but the way is set up it is a core part of Pineboo now.
# ... we should probably create our own one. Not this one.


class FLManager(QtCore.QObject, IManager):
    """
    Serve as the database administrator.

    Responsible for opening the forms or obtaining their definitions (.ui files).
    It also maintains the metadata of all tables in the database base.
    data, offering the ability to retrieve metadata
    using PNTableMetaData objects from a given table.

    @author InfoSiAL S.L.
    """

    init_count_: int = 0  # Indica el número de veces que se ha llamado a FLManager::init()

    def __init__(self, db: "iconnection.IConnection") -> None:
        """
        Inicialize.
        """
        super().__init__()
        self.db_ = db

        self.list_tables_ = []
        self.dict_key_metadata_ = {}
        self.cache_metadata_ = {}
        self._cache_action = {}
        QtCore.QTimer.singleShot(100, self.init)
        self.init_count_ += 1
        self._util = flutil.FLUtil()

    def init(self) -> None:
        """
        Initialize.
        """
        self.init_count_ = self.init_count_ + 1
        # self.createSystemTable("flmetadata")
        # self.createSystemTable("flseqs")

        if not self.db_:
            raise Exception("FLManagar.__init__. self.db_ is empty!")

        if not self.db_.connManager().dbAux():
            return

        if not self.cache_metadata_:
            self.cache_metadata_ = {}

        if not self._cache_action:
            self._cache_action = {}

    def finish(self) -> None:
        """Apply close process."""

        self.dict_key_metadata_ = {}
        self.list_tables_ = []
        self.cache_metadata_ = {}
        self._cache_action = {}

    def metadata(
        self,
        metadata_name_or_xml: Union[str, "ElementTree.Element", Dict[str, Any]],
        quick: bool = False,
    ) -> Optional["pntablemetadata.PNTableMetaData"]:
        """
        To obtain definition of a database table, from an XML file.

        The name of the table corresponds to the name of the file plus the extension ".mtd"
        which contains in XML the description of the tables. This method scans the file
        and builds / returns the corresponding PNTableMetaData object, in addition
        make a copy of these metadata in a table of the same database
        to be able to determine when it has been modified and thus, if necessary, rebuild
        the table so that it adapts to the new metadata. NOT MADE
        CHECKS OF SYNTATIC ERRORS IN THE XML.

        IMPORTANT: For these methods to work properly, it is strictly
            it is necessary to have created the database in PostgreSQL with coding
            UNICODE; "createdb -E UNICODE abanq".

        @param metadata_name_or_xml Name of the database table from which to obtain the metadata
        @param quick If TRUE does not check, use carefully
        @return A PNTableMetaData object with the metadata of the requested table
        """

        if metadata_name_or_xml in (None, ""):
            return None

        if not self.db_:
            raise Exception("metadata. self.db_ is empty!")

        if not quick:
            quick = not application.PROJECT.db_admin_mode

        # if quick is None:
        #    dbadmin = settings.CONFIG.value("application/dbadmin_enabled", False)
        #    quick = not bool(dbadmin)

        if isinstance(metadata_name_or_xml, str):
            ret: Optional["pntablemetadata.PNTableMetaData"] = None
            acl: Optional["pnaccesscontrollists.PNAccessControlLists"] = None
            key = metadata_name_or_xml.strip()
            stream = None

            table_name = key if key.endswith(".mtd") else "%s.mtd" % key

            if table_name in self.cache_metadata_.keys():
                ret = copy.copy(self.cache_metadata_[table_name])
            if not ret:
                # Buscamos primero si es un model_.
                model_name = (
                    metadata_name_or_xml[: metadata_name_or_xml.find(".mtd")]
                    if metadata_name_or_xml.find(".mtd") > -1
                    else metadata_name_or_xml
                )
                model_ = qsadictmodules.QSADictModules.from_project("%s_orm" % model_name)
                if model_ is not None:
                    ret = self.metadata(model_.legacy_metadata, quick)
                else:  # extraer datos del mtd (modo clásico)
                    # LOGGER.warning("metadata %s from xml is deprecated", metadata_name_or_xml)
                    stream = self.db_.connManager().managerModules().contentCached(table_name)

                    if not stream:
                        if metadata_name_or_xml.find("alteredtable") == -1:
                            LOGGER.info(
                                "FLManager : "
                                + self._util.translate(
                                    "FLManager",
                                    "Error al cargar los metadatos para la tabla %s"
                                    % metadata_name_or_xml,
                                )
                            )
                        return None

                    # docElem = doc.documentElement()
                    # tree = utils_base.load2xml(stream)
                    tree = None
                    try:
                        tree = ElementTree.fromstring(stream)

                    except Exception as error:
                        LOGGER.error("Failed loading %s :%s", metadata_name_or_xml, error)

                    if tree is None:
                        return None

                    ret = self.metadata(tree, quick)

                if ret is None:
                    return None

                if ret.isQuery():
                    ret.setName(model_name)

                if not quick:
                    self.cache_metadata_[table_name] = copy.copy(ret)

                    if not self.existsTable(metadata_name_or_xml):
                        self.createTable(ret)
                    else:
                        if self.db_.mismatchedTable(metadata_name_or_xml, ret):
                            if ret.name():
                                msg = self._util.translate(
                                    "application",
                                    "La estructura de los metadatos de la %s '%s'"
                                    " y su estructura interna en la base de datos no coinciden.\n"
                                    "Regenerando la base de datos."
                                    % ("vista" if ret.isQuery() else "tabla", metadata_name_or_xml),
                                )
                                LOGGER.warning(msg)

                                # must_alter = self.db_.mismatchedTable(metadata_name_or_xml, ret)
                                # if must_alter:
                                # if not self.alterTable(stream, stream, "", True):
                                if not self.alterTable(ret):
                                    LOGGER.warning(
                                        "La regeneración de la %s %s ha fallado",
                                        "vista" if ret.isQuery() else "tabla",
                                        metadata_name_or_xml,
                                    )
                            else:
                                LOGGER.warning(
                                    "El metadata %s informa como una tabla algo que no es.",
                                    metadata_name_or_xml,
                                )

                # throwMsgWarning(self.db_, msg)

            acl = application.PROJECT.aq_app.acl()

            if acl:
                acl.process(ret)

            return ret

        elif isinstance(metadata_name_or_xml, dict):
            meta_dict = metadata_name_or_xml

            name = meta_dict["name"]
            alias = meta_dict["alias"]
            query = meta_dict["query"] if "query" in meta_dict.keys() else ""
            # visible = True
            # editable = True
            ftsfun = meta_dict["ftsfunction"] if "ftsfunction" in meta_dict.keys() else ""
            concur_warn = meta_dict["concurwarn"] if "concurwarn" in meta_dict.keys() else False
            detect_locks = meta_dict["detectlocks"] if "detectlocks" in meta_dict.keys() else False
            cached_fields = meta_dict["cachedfields"] if "cachedfields" in meta_dict.keys() else ""

            table_metadata = pntablemetadata.PNTableMetaData(name, alias, query)
            table_metadata.setFTSFunction(ftsfun)
            table_metadata.setConcurWarn(concur_warn)
            table_metadata.setDetectLocks(detect_locks)
            table_metadata.setCachedFields(cached_fields)

            compound_key = pncompoundkeymetadata.PNCompoundKeyMetaData()
            assocs = []

            for child in meta_dict["fields"]:
                field_mtd = self.metadataField(child)
                table_metadata.addFieldMD(field_mtd)
                if field_mtd.isCompoundKey():
                    compound_key.addFieldMD(field_mtd)

                if field_mtd.associatedFieldName():
                    assocs.append(
                        [
                            field_mtd.associatedFieldName(),
                            field_mtd.associatedFieldFilterTo(),
                            field_mtd.name(),
                        ]
                    )

        else:
            # QDomDoc
            # root = n.getroot()
            name: str = ""  # type: ignore [no-redef] # noqa: F821
            query: str = ""  # type: ignore [no-redef] # noqa: F821
            alias: str = ""  # type: ignore [no-redef] # noqa: F821
            ftsfun: str = ""  # type: ignore [no-redef] # noqa: F821
            visible = True
            editable = True
            concur_warn = False
            detect_locks = False
            cached_fields = ""
            child_list: List["ElementTree.Element"] = []
            for child in metadata_name_or_xml:
                if child.tag == "field":
                    child_list.append(child)
                elif child.tag == "name":
                    name = child.text or ""
                elif child.tag == "query":
                    query = child.text or ""
                elif child.tag == "alias":
                    alias = self._util.translate(
                        "Metadata", utils_base.auto_qt_translate_text(child.text)
                    )
                elif child.tag == "visible":
                    visible = child.text == "true"
                elif child.tag == "editable":
                    editable = child.text == "true"
                elif child.tag == "detectLocks":
                    detect_locks = child.text == "true"
                elif child.tag == "concurWarn":
                    concur_warn = child.text == "true"
                elif child.tag == "FTSFunction":
                    ftsfun = child.text or ""
                elif child.tag == "cachedfields":
                    cached_fields = child.text or ""
                    if cached_fields:
                        LOGGER.warning("MTD(%s): cachedfields found :%s" % (name, cached_fields))

            table_metadata = pntablemetadata.PNTableMetaData(name, alias, query)

            table_metadata.setFTSFunction(ftsfun)
            table_metadata.setConcurWarn(concur_warn)
            table_metadata.setDetectLocks(detect_locks)
            table_metadata.setCachedFields(cached_fields)

            compound_key = pncompoundkeymetadata.PNCompoundKeyMetaData()
            assocs = []

            for child in child_list:
                if child.tag == "field":
                    field_mtd = self.metadataField(child, visible, editable)
                    table_metadata.addFieldMD(field_mtd)
                    if field_mtd.isCompoundKey():
                        compound_key.addFieldMD(field_mtd)

                    if field_mtd.associatedFieldName():
                        assocs.append(
                            [
                                field_mtd.associatedFieldName(),
                                field_mtd.associatedFieldFilterTo(),
                                field_mtd.name(),
                            ]
                        )

        table_metadata.setCompoundKey(compound_key)

        for assoc_with, assoc_by, field_name in assocs:
            field_metadata = table_metadata.field(field_name or "")
            assoc_field_mtd = table_metadata.field(assoc_with or "")
            if field_metadata and assoc_field_mtd and assoc_by:
                field_metadata.setAssociatedField(assoc_field_mtd, assoc_by)

        acl = application.PROJECT.aq_app.acl()
        if acl:
            acl.process(table_metadata)

        return table_metadata

    @decorators.not_implemented_warn
    def metadataDev(self, table_name: str, quick: bool = False) -> bool:
        """Return metadata (deprecated)."""
        return True

    def query(
        self, name: str, parent: Optional["pnsqlquery.PNSqlQuery"] = None
    ) -> Optional["pnsqlquery.PNSqlQuery"]:
        """
        To obtain a query of the database, from an XML file.

        The name of the query corresponds to the name of the file plus the extension ".qry"
        which contains the description of the query in XML. This method scans the file
        and build / return the FLSqlQuery object. NOT MADE
        CHECKS OF SYNTATIC ERRORS IN THE XML.

        @param name Name of the database query you want to obtain
        @return An FLSqlQuery object that represents the query you want to get
        """
        if self.db_ is None:
            raise Exception("query. self.db_ is empty!")

        qry_ = self.db_.connManager().managerModules().contentCached("%s.qry" % name)

        if not qry_:
            return None

        qry = pnsqlquery.PNSqlQuery()
        qry.setName(name)
        root_ = ElementTree.fromstring(qry_)

        for child in root_:
            text_ = child.text or ""
            value = (
                text_.replace("\t", "").replace("\n", " ").replace("\r", "").strip()
                if child.tag in ["select", "from", "tables", "order"]
                else ""
            )
            if child.tag == "select":
                qry.setSelect(value)
            elif child.tag == "from":
                qry.setFrom(value)
            elif child.tag == "tables":
                qry.setTablesList(value)
            elif child.tag == "order":
                qry.setOrderBy(value)
            elif child.tag == "where":
                for child_where in root_.iter("where"):
                    text_ = child_where.text or ""
                    value = text_.replace("\t", "").replace("\n", "").replace("\r", "").strip()
                    qry.setWhere(value)
            elif child.tag == "group":
                for level, child_group in enumerate(root_.findall("group")):
                    child_level = child_group.find("level")
                    child_field = child_group.find("field")
                    if hasattr(child_field, "text") and hasattr(child_level, "text"):
                        text_level = getattr(child_level, "text", "")
                        if level == int(
                            text_level.replace("\t", "").replace("\n", "").replace("\r", "").strip()
                        ):
                            text_value_level = (
                                getattr(child_field, "text", "")
                                .replace("\t", "")
                                .replace("\n", "")
                                .replace("\r", "")
                                .strip()
                            )
                            qry.addGroup(pngroupbyquery.PNGroupByQuery(level, text_value_level))

        return qry

    def action(self, action_name: str) -> "pnaction.PNAction":
        """
        Return the definition of an action from its name.

        This method looks in the [id_modulo] .xml for the action that is passed
        as a name and build and return the corresponding FLAction object.
        NO SYMPATHIC ERROR CHECKS ARE MADE IN THE XML.

        @param n Name of the action
        @return A FLAction object with the description of the action

        """

        if not self.db_:
            raise Exception("action. self.db_ is empty!")

        # FIXME: This function is really inefficient. Pineboo already parses the actions much before.
        if action_name in self._cache_action.keys():
            pnaction_ = self._cache_action[action_name]
        else:
            pnaction_ = convert_flaction.convert_to_flaction(action_name)
            if not pnaction_.table():
                pnaction_.setTable(action_name)

            self._cache_action[action_name] = pnaction_

        return pnaction_

    def existsTable(self, table_name: str, cache: bool = True) -> bool:
        """
        Check if the table specified in the database exists.

        @param n Name of the table to check if it exists
        @param cache If a certain query first checks the table cache, otherwise
                    make a query to the base to obtain the existing tables
        @return TRUE if the table exists, FALSE otherwise.
        """
        if not self.db_ or table_name is None:
            return False

        if cache and table_name in self.list_tables_:
            return True
        else:
            ret = (
                self.db_.connManager()
                .dbAux()
                .existsTable(table_name if not table_name.endswith(".mtd") else table_name[:-4])
            )
            if ret:
                self.list_tables_.append(table_name)

            return ret

    def checkMetaData(
        self,
        metadata_or_name: Union[str, "pntablemetadata.PNTableMetaData"],
        metadata2: "pntablemetadata.PNTableMetaData",
    ) -> bool:
        """
        Compare the metadata of two tables.

        The XML definition of those two tables is they pass as two strings of characters.

        @param mtd1 Character string with XML describing the first table
        @param mtd2 Character string with XML describing the second table
        @return TRUE if the two descriptions are equal, and FALSE otherwise
        """
        if isinstance(metadata_or_name, str):
            return metadata_or_name == metadata2.name()
        else:
            if None in [metadata_or_name, metadata2] or len(metadata_or_name.fieldList()) != len(
                metadata2.fieldList()
            ):
                return False

            for field1 in [field for field in metadata_or_name.fieldList() if not field.isCheck()]:
                field2 = metadata2.field(field1.name())
                if field2 is None:
                    return False

                if field2.isCheck():
                    continue

                if (
                    field1.type() != field2.type()
                    or field1.allowNull() != field2.allowNull()
                    or field1.isUnique() != field2.isUnique()
                    or field1.isIndex() != field2.isIndex()
                    or field1.length() != field2.length()
                    or field1.partDecimal() != field2.partDecimal()
                    or field1.partInteger() != field2.partInteger()
                ):
                    return False

            return True

    def alterTable(self, new_metadata: "pntablemetadata.PNTableMetaData") -> bool:
        """
        Modify the structure or metadata of a table. Preserving the possible data that can contain.

        According to the existing definition of metadata in the .mtd file at a given time, this
        method reconstructs the table with those metadata without the loss of information or data,
        that might exist at that time in the table.

        @param n Name of the table to rebuild
        @param mtd1 XML description of the old structure
        @param mtd2 XML description of the new structure
        @param key Sha1 key of the old structure
        @return TRUE if the modification was successful
        """
        if not self.db_:
            raise Exception("alterTable. self.db_ is empty!")

        return self.db_.alterTable(new_metadata)

    def createTable(
        self, metadata_or_name: Optional[Union[str, "pntablemetadata.PNTableMetaData"]] = None
    ) -> Optional["pntablemetadata.PNTableMetaData"]:
        """
        Create a table in the database.

        @param n_tmd Name or metadata of the table you want to create
        @return A PNTableMetaData object with the metadata of the table that was created, or
          0 if the table could not be created or already existed
        """
        if not self.db_:
            raise Exception("createTable. self.db_ is empty!")

        if isinstance(metadata_or_name, str):
            metadata_or_name = self.metadata(metadata_or_name)

        if metadata_or_name is None:
            return None

        if metadata_or_name.name() not in self.list_tables_:
            if not self.existsTable(metadata_or_name.name()):
                if not self.db_.connManager().default().createTable(metadata_or_name):
                    LOGGER.warning(
                        "createTable: %s",
                        self.tr(
                            "%s %s could not be created"
                            % (
                                "view" if metadata_or_name.isQuery() else "table",
                                metadata_or_name.name(),
                            )
                        ),
                    )
                    return None

            self.list_tables_.append(metadata_or_name.name())

        return metadata_or_name

    def formatValueLike(
        self,
        fmd_or_type: Union["pnfieldmetadata.PNFieldMetaData", str],
        value: Any,
        upper: bool = False,
    ) -> str:
        """
        Return the value content of a formatted field.

        Return the value content of a formatted field to be recognized by the current database in LIKE conditions.
        within the SQL WHERE closing.

        This method takes as parameters the field metadata defined with
        PNFieldMetaData. In addition to TRUE and FALSE as possible values of a field
        logical also accepts the values Yes and No (or its translation into the corresponding language).
        The dates are adapted to the YYYY-MM-DD form, which is the format recognized by PostgreSQL.

        @param fmd_or_type PNFieldMetaData object that describes the metadata for the field
        @param value Value to be formatted for the indicated field
        @param upper If TRUE converts the value to uppercase (if it is a string type)
        """

        if not self.db_:
            raise Exception("formatValueLike. self.db_ is empty!")

        return self.db_.formatValueLike(
            fmd_or_type if isinstance(fmd_or_type, str) else fmd_or_type.type(), value, upper
        )

    def formatAssignValueLike(self, *args) -> str:
        """
        Return the value content of a formatted field to be recognized by the current database, within the SQL WHERE closing.

        This method takes as parameters the field metadata defined with
        PNFieldMetaData. In addition to TRUE and FALSE as possible values of a field
        logical also accepts the values Yes and No (or its translation into the corresponding language).
        The dates are adapted to the YYYY-MM-DD form, which is the format recognized by PostgreSQL.

        @param field_metadata PNFieldMetaData object that describes the metadata for the field
        @param v Value to be formatted for the indicated field
        @param upper If TRUE converts the value to uppercase (if it is a string type)
        """

        type_: str = ""
        field_name_: str = ""
        args_0: Union[pnfieldmetadata.PNFieldMetaData, str] = args[0]
        value_: Any = args[1]
        upper_: bool = args[2]

        if isinstance(args_0, pnfieldmetadata.PNFieldMetaData):
            field_name_ = args_0.name()
            type_ = args_0.type()

            mtd_ = args_0.metadata()

            if mtd_ is None:
                return ""

            if args_0.isPrimaryKey():
                field_name_ = mtd_.primaryKey(True)

            if mtd_.isQuery() and field_name_.find(".") == -1:
                qry = pnsqlquery.PNSqlQuery(mtd_.query())

                for item in qry.fieldList():
                    item_field_name = item[item.find(".") + 1 :] if item.find(".") > -1 else item
                    if item_field_name == field_name_:
                        break
            else:
                item_field_name = args_0.name()

            field_name_ = "%s.%s" % (mtd_.name(), item_field_name)

        else:
            field_name_ = args_0
            type_ = args[1] if isinstance(args[1], str) else args[1].type()

            if len(args) == 4:
                value_ = args[2]
                upper_ = args[3]

        format_value_ = self.formatValueLike(type_, value_, upper_)

        if not format_value_:
            return "1 = 1"

        field_name_ = (
            "upper(%s)" % field_name_
            if upper_ and type_ in ["string", "stringlist", "timestamp"]
            else field_name_
        )

        return "%s %s" % (field_name_, format_value_)

    def formatValue(self, field_metadata_or_type: str, v: Any, upper: bool = False) -> str:
        """Return format value."""

        if not self.db_:
            raise Exception("formatValue. self.db_ is empty!")

        if not field_metadata_or_type:
            raise ValueError("field_metadata_or_type is required")

        if not isinstance(field_metadata_or_type, str):
            field_metadata_or_type = field_metadata_or_type.type()

        return str(self.db_.formatValue(field_metadata_or_type, v, upper))

    def formatAssignValue(self, *args) -> str:
        """Return format assign value."""

        if args[0] is None:
            # print("FLManager.formatAssignValue(). Primer argumento vacio %s" % args[0])
            return "1 = 1"

        field_name_: str = ""
        field_type_: str = ""
        value_: Any = None
        upper_: bool = False

        if isinstance(args[0], pnfieldmetadata.PNFieldMetaData):
            if len(args) == 3:
                field_metadata = args[0]
                mtd = field_metadata.metadata()
                if mtd is None:
                    field_name_ = field_metadata.name()
                    field_type_ = field_metadata.type()

                elif field_metadata.isPrimaryKey():
                    field_name_ = mtd.primaryKey(True)
                    field_type_ = field_metadata.type()

                else:
                    field_name_ = field_metadata.name()
                    if mtd.isQuery() and "." not in field_name_:
                        prefix_table_ = mtd.name()
                        qry = self.query(mtd.query())

                        if qry:
                            for field in qry.fieldList():
                                field_section_ = field
                                pos = field.find(".")
                                if pos > -1:
                                    prefix_table_ = field[:pos]
                                    field_section_ = field[pos + 1 :]
                                else:
                                    field_section_ = field

                                if field_section_ == field_name_:
                                    break

                        field_name_ = "%s.%s" % (prefix_table_, field_name_)

                    field_type_ = args[0].type()

                value_ = args[1]
                upper_ = args[2]

            elif len(args) == 2:
                field_name_ = args[0].name()
                field_type_ = args[0].type()
                value_ = args[1]

        elif isinstance(args[0], str):
            field_name_ = args[0]
            field_type_ = (
                args[1].type() if isinstance(args[1], pnfieldmetadata.PNFieldMetaData) else args[1]
            )
            value_ = args[2]
            upper_ = args[3]

        if not field_type_:
            return "1 = 1"

        format_value_ = self.formatValue(field_type_, value_, upper_)
        if format_value_ is None:
            return "1 = 1"

        if not field_name_:
            field_name_ = args[0] if isinstance(args[0], str) else args[0].name()

        field_name_ = (
            "upper(%s)" % field_name_
            if upper_ and field_type_ in ["string", "stringlist", "timestamp"]
            else field_name_
        )

        return "%s %s %s" % (
            field_name_,
            "LIKE" if field_type_ == "string" and format_value_.find("%") > -1 else "=",
            format_value_,
        )

    def metadataField(
        self, field: Union["ElementTree.Element", Dict], v: bool = True, ed: bool = True
    ) -> "pnfieldmetadata.PNFieldMetaData":
        """
        Create a PNFieldMetaData object from an XML element.

        Given an XML element, which contains the description of a
        table field builds and adds to a list of descriptions
        of fields the corresponding PNFieldMetaData object, which contains
        said definition of the field. It also adds it to a list of keys
        composite, if the constructed field belongs to a composite key.
        NO SYMPATHIC ERROR CHECKS ARE MADE IN THE XML.

        @param field XML element with field description
        @param v Default value used for the visible property
        @param ed Value used by default for editable property
        @return PNFieldMetaData object that contains the description of the field
        """
        if field in (None, ""):  # type: ignore [comparison-overlap]
            raise ValueError("field is required")

        valid_types = [
            "int",
            "uint",
            "bool",
            "double",
            "time",
            "date",
            "pixmap",
            "bytearray",
            "string",
            "stringlist",
            "unlock",
            "serial",
            "timestamp",
            "json",
        ]

        compound_key = False
        name: str = ""
        alias: str = ""
        options_list: Optional[str] = None
        reg_exp = None
        search_options = None

        as_null = True
        is_primary_key = False
        calculated = False
        is_index = False
        is_unique = False
        coun = False
        out_transaction = False
        visible_grid = True
        full_calculated = False
        trimm = False
        visible = True
        editable = True

        type_: Optional[str] = None
        length = 0
        part_integer = 4
        part_decimal = 0

        default_value = None
        if isinstance(field, dict):
            for tag in field.keys():
                if tag in ("relation", "associated"):
                    continue

                elif tag == "name":
                    name = field[tag]

                elif tag == "alias":
                    alias = utils_base.auto_qt_translate_text(field[tag])

                elif tag == "null":
                    as_null = field[tag]

                elif tag == "pk":
                    is_primary_key = field[tag]

                elif tag == "type":
                    type_ = field[tag]

                elif tag == "length":
                    length = field[tag]

                elif tag == "regexp":
                    reg_exp = field[tag]

                elif tag == "default":
                    default_value = field[tag]
                    if isinstance(default_value, str):
                        default_value = utils_base.auto_qt_translate_text(default_value)

                elif tag == "outtransaction":
                    out_transaction = field[tag]

                elif tag == "counter":
                    coun = field[tag]

                elif tag == "calculated":
                    calculated = field[tag]

                elif tag == "fullycalculated":
                    full_calculated = field[tag]

                elif tag == "trimmed":
                    trimm = field[tag]

                elif tag == "visible":
                    visible = field[tag]

                elif tag == "visiblegrid":
                    visible_grid = field[tag]

                elif tag == "editable":
                    editable = field[tag]

                elif tag == "partI":
                    part_integer = field[tag]

                elif tag == "partD":
                    part_decimal = field[tag]

                elif tag == "index":
                    is_index = field[tag]

                elif tag == "unique":
                    is_unique = field[tag]

                elif tag == "ck":
                    compound_key = field[tag]

                elif tag == "optionslist":
                    options_list = ", ".join(field[tag])

                elif tag == "searchoptions":
                    search_options = ", ".join(field[tag])

        else:
            for child in field:
                tag = child.tag
                if tag in ("relation", "associated"):
                    continue

                elif tag == "name":
                    name = child.text or ""
                    continue

                elif tag == "alias":
                    alias = utils_base.auto_qt_translate_text(child.text)
                    continue

                elif tag == "null":
                    as_null = child.text == "true"
                    continue

                elif tag == "pk":
                    is_primary_key = child.text == "true"
                    continue

                elif tag == "type":
                    type_ = child.text
                    if type_ not in valid_types:
                        type_ = None
                    continue

                elif tag == "length":
                    length = int(child.text or 0)
                    continue

                elif tag == "regexp":
                    reg_exp = child.text
                    continue

                elif tag == "default":
                    default_value = child.text or ""
                    if default_value.find("QT_TRANSLATE_NOOP") > -1:
                        default_value = utils_base.auto_qt_translate_text(default_value)
                    continue
                elif tag == "outtransaction":
                    out_transaction = child.text == "true"
                    continue
                elif tag == "counter":
                    coun = child.text == "true"
                    continue
                elif tag == "calculated":
                    calculated = child.text == "true"
                    continue
                elif tag == "fullycalculated":
                    full_calculated = child.text == "true"
                    continue
                elif tag == "trimmed":
                    trimm = child.text == "true"
                    continue
                elif tag == "visible":
                    visible = child.text == "true"
                    continue
                elif tag == "visiblegrid":
                    visible_grid = child.text == "true"
                    continue
                elif tag == "editable":
                    editable = child.text == "true"
                    continue
                elif tag == "partI":
                    part_integer = int(child.text or 4)
                    continue
                elif tag == "partD":
                    part_decimal = int(child.text or 0)
                    continue
                elif tag == "index":
                    is_index = child.text == "true"
                    continue
                elif tag == "unique":
                    is_unique = child.text == "true"
                    continue
                elif tag == "ck":
                    compound_key = child.text == "true"
                    continue
                elif tag == "optionslist":
                    options_list = child.text
                    continue
                elif tag == "searchoptions":
                    search_options = child.text
                    continue

        field_mtd = pnfieldmetadata.PNFieldMetaData(
            name,
            self._util.translate("Metadata", alias),
            as_null,
            is_primary_key,
            type_,
            length,
            calculated,
            visible,
            editable,
            part_integer,
            part_decimal,
            is_index,
            is_unique,
            coun,
            default_value,
            out_transaction,
            reg_exp,
            visible_grid,
            True,
            compound_key,
        )
        field_mtd.setFullyCalculated(full_calculated)
        field_mtd.setTrimed(trimm)

        if options_list:
            field_mtd.setOptionsList(options_list)
        if search_options:
            field_mtd.setSearchOptions(search_options)

        assoc_by = ""
        assoc_with = ""
        if isinstance(field, dict):
            for tag in field.keys():
                if tag == "relations":
                    for rel_ in field[tag]:
                        field_mtd.addRelationMD(self.metadataRelation(rel_))

                elif tag == "associated":
                    for sub_tag in field[tag].keys():
                        if sub_tag == "with":
                            assoc_with = field[tag][sub_tag]
                        elif sub_tag == "by":
                            assoc_by = field[tag][sub_tag]

            if assoc_with and assoc_by:
                field_mtd.setAssociatedField(assoc_with, assoc_by)

        else:
            for child in field:
                tag = child.tag
                if tag == "relation":
                    field_mtd.addRelationMD(self.metadataRelation(child))
                    continue
                elif tag == "associated":
                    for sub in child:
                        sub_tag = sub.tag
                        if sub_tag == "with":
                            assoc_with = sub.text or ""
                        elif sub_tag == "by":
                            assoc_by = sub.text or ""

            if assoc_with and assoc_by:
                field_mtd.setAssociatedField(assoc_with, assoc_by)

        return field_mtd

    def metadataRelation(
        self, relation: Union[QtXml.QDomElement, "ElementTree.Element"]
    ) -> "pnrelationmetadata.PNRelationMetaData":
        """
        Create a FLRelationMetaData object from an XML element.

        Given an XML element, which contains the description of a
        relationship between tables, builds and returns the FLRelationMetaData object
        corresponding, which contains said definition of the relationship.
        NO SYMPATHIC ERROR CHECKS ARE MADE IN THE XML.

        @param relation XML element with the description of the relationship
        @return FLRelationMetaData object that contains the description of the relationship
        """
        if relation in (None, ""):
            raise ValueError("relation is required")

        foreign_table = ""
        foreign_field = ""
        relation_card = pnrelationmetadata.PNRelationMetaData.RELATION_M1
        delete_cascade = False
        update_cascade = False
        check_integrity = True

        if isinstance(relation, QtXml.QDomElement):
            LOGGER.warning(
                "QtXml.QDomElement is deprecated for declare metadata relation. Please use xml.etree.ElementTree.ElementTree instead"
            )
            node = relation.firstChild()

            while not node.isNull():
                elem = node.toElement()
                if not elem.isNull():
                    if elem.tagName() == "table":
                        foreign_table = elem.text()
                        node = node.nextSibling()
                        continue

                    elif elem.tagName() == "field":
                        foreign_field = elem.text()
                        node = node.nextSibling()
                        continue

                    elif elem.tagName() == "card":
                        if elem.text() == "1M":
                            relation_card = pnrelationmetadata.PNRelationMetaData.RELATION_1M

                        node = node.nextSibling()
                        continue

                    elif elem.tagName() == "delC":
                        delete_cascade = elem.text() == "true"
                        node = node.nextSibling()
                        continue

                    elif elem.tagName() == "updC":
                        update_cascade = elem.text() == "true"
                        node = node.nextSibling()
                        continue

                    elif elem.tagName() == "checkIn":
                        check_integrity = elem.text() == "true"
                        node = node.nextSibling()
                        continue

                node = node.nextSibling()

        elif isinstance(relation, dict):
            for key in relation.keys():
                if key == "table":
                    foreign_table = relation[key]
                elif key == "field":
                    foreign_field = relation[key]
                elif key == "card":
                    if relation[key] == "1M":
                        relation_card = pnrelationmetadata.PNRelationMetaData.RELATION_1M
                elif key == "delC":
                    delete_cascade = relation[key]
                elif key == "updC":
                    update_cascade = relation[key]
                elif key == "checkIn":
                    check_integrity = relation[key]
        else:
            for child in relation:
                tag = child.tag
                if tag == "table":
                    foreign_table = child.text or ""
                    continue
                elif tag == "field":
                    foreign_field = child.text or ""
                    continue
                elif tag == "card":
                    if child.text == "1M":
                        relation_card = pnrelationmetadata.PNRelationMetaData.RELATION_1M
                    continue
                elif tag == "delC":
                    delete_cascade = child.text == "true"
                    continue
                elif tag == "updC":
                    update_cascade = child.text == "true"
                    continue
                elif tag == "checkIn":
                    check_integrity = child.text == "true"
                    continue

        return pnrelationmetadata.PNRelationMetaData(
            foreign_table,
            foreign_field,
            relation_card,
            delete_cascade,
            update_cascade,
            check_integrity,
        )

    @decorators.not_implemented_warn
    def queryParameter(self, parameter: QtXml.QDomElement) -> Any:
        """
        Create an FLParameterQuery object from an XML element.

        Given an XML element, which contains the description of a
        parameter of a query, build and return the FLParameterQuery object
        correspondent.
        NO SYMPATHIC ERROR CHECKS ARE MADE IN THE XML.

        @param parameter XML element with the description of the parameter of a query
        @return FLParameterQuery object that contains the parameter description.
        """
        return True

    @decorators.not_implemented_warn
    def queryGroup(self, group: QtXml.QDomElement) -> Any:
        """
        Create a FLGroupByQuery object from an XML element.

        Given an XML element, which contains the description of a grouping level
        of a query, build and return the corresponding FLGroupByQuery object.
        NO SYMPATHIC ERROR CHECKS ARE MADE IN THE XML.

        @param group XML element with the description of the grouping level of a query.
        @return FLGroupByQuery object that contains the description of the grouping level
        """
        return True

    def createSystemTable(self, table_name: str) -> bool:
        """
        Create a system table.

        This method reads directly from disk the file with the description of a table
        of the system and creates it in the database. Its normal use is to initialize
        The system with initial tables.

        @param table_name Name of the table.
        @return A PNTableMetaData object with the metadata of the table that was created, or
          False if the table could not be created or already existed.
        """

        return self.createTable(self.metadata("%s.mtd" % table_name, True)) is not None

    def loadTables(self) -> None:
        """
        Load in the table list the names of the database tables.
        """
        if not self.db_:
            raise Exception("loadTables. self.db_ is empty!")

        self.list_tables_.clear()

        self.list_tables_ = self.db_.connManager().dbAux().tables()

    def cleanupMetaData(self) -> None:
        """
        Clean the flmetadata table, update the xml content with that of the .mtd file currently loaded.
        """
        if not self.db_:
            raise Exception("cleanupMetaData. self.db_ is empty!")

        if not self.existsTable("flfiles") or not self.existsTable("flmetadata"):
            return

        buffer = None
        table = ""

        # q.setForwardOnly(True)
        # c.setForwardOnly(True)

        if self.dict_key_metadata_:
            self.dict_key_metadata_.clear()
        else:
            self.dict_key_metadata_ = {}

        self.loadTables()
        self.db_.connManager().managerModules().loadKeyFiles()
        self.db_.connManager().managerModules().loadAllIdModules()
        self.db_.connManager().managerModules().loadIdAreas()

        qry = pnsqlquery.PNSqlQuery(None, "dbAux")
        qry.exec_("SELECT tabla,xml FROM flmetadata")
        while qry.next():
            self.dict_key_metadata_[str(qry.value(0))] = str(qry.value(1))

        cursor = pnsqlcursor.PNSqlCursor("flmetadata", True, "dbAux")

        qry2 = pnsqlquery.PNSqlQuery(None, "dbAux")
        qry2.exec_(
            "SELECT nombre,sha FROM flfiles WHERE nombre LIKE '%.mtd' and nombre not like '%%alteredtable%'"
        )
        while qry2.next():
            table = str(qry2.value(0))
            table = table.replace(".mtd", "")
            tmd = self.metadata(table, True)
            if not self.existsTable(table):
                self.createTable(table)
            if not tmd:
                LOGGER.warning(
                    "FLManager::cleanupMetaData %s",
                    self._util.translate(
                        "application", "No se ha podido crear los metadatatos para la tabla %s"
                    )
                    % table,
                )
                continue

            cursor.select("tabla = '%s'" % table)
            if cursor.next():
                buffer = cursor.prime_update()
                buffer.set_value("xml", qry2.value(1))
                cursor.update()
            self.dict_key_metadata_[table] = qry2.value(1)

    @classmethod
    def isSystemTable(cls, table_name: str) -> bool:
        """
        Return if the given table is a system table.

        @param n Name of the table.
        @return TRUE if it is a system table
        """
        return table_name[0:3] == "sys" or table_name.startswith(
            (
                "flfiles",
                "flmetadata",
                "flmodules",
                "flareas",
                "flserial",
                "flvar",
                "flsettings",
                "flseqs",
                "flupdates",
                "flacls",
                "flacos",
                "flacs",
                "flgroups",
                "flusers",
            )
        )

    def storeLargeValue(
        self, mtd: "pntablemetadata.PNTableMetaData", large_value: str
    ) -> Optional[str]:
        """
        Store large field values in separate indexed tables by SHA keys of the value content.

        It is used to optimize queries that include fields with large values,
        such as images, to handle the reference to the value in the SQL queries
        which is of constant size instead of the value itself. This decreases traffic by
        considerably reduce the size of the records obtained.

        Queries can use a reference and get its value only when
        you need via FLManager :: fetchLargeValue ().


        @param mtd Metadata of the table containing the field
        @param large_value Large field value
        @return Value reference key
        """
        if not self.db_:
            raise Exception("storeLareValue. self.db_ is empty!")

        if large_value[0:3] == "RK@":
            return None

        table_name = mtd.name()

        table_large = "fllarge"

        if not application.PROJECT.aq_app.singleFLLarge():
            table_large = "fllarge_%s" % table_name

            if not self.existsTable(table_large):
                mtd_large = pntablemetadata.PNTableMetaData(table_large, table_large)
                field_refkey = pnfieldmetadata.PNFieldMetaData(
                    "refkey", "refkey", False, True, "string", 100
                )

                field_sha = pnfieldmetadata.PNFieldMetaData("sha", "sha", True, False, "string", 50)

                field_contenido = pnfieldmetadata.PNFieldMetaData(
                    "contenido", "contenido", True, False, "stringlist"
                )

                mtd_large.addFieldMD(field_refkey)
                mtd_large.addFieldMD(field_sha)
                mtd_large.addFieldMD(field_contenido)

                if not self.createTable(mtd_large):
                    return None

        sha = str(self._util.sha1(large_value))
        ref_key = "RK@%s@%s" % (table_name, sha)
        qry = pnsqlquery.PNSqlQuery(None, "dbAux")
        qry.setSelect("refkey")
        qry.setFrom(table_large)
        qry.setWhere("refkey = '%s'" % ref_key)
        if qry.exec_() and qry.first():
            if qry.value(0) != sha:
                sql = "UPDATE %s SET contenido = '%s' WHERE refkey ='%s'" % (
                    table_large,
                    large_value,
                    ref_key,
                )

        else:
            sql = "INSERT INTO %s (contenido,refkey) VALUES ('%s','%s')" % (
                table_large,
                large_value,
                ref_key,
            )

        self.db_.connManager().useConn("dbAux").execute_query(sql)
        return ref_key

    def fetchLargeValue(self, ref_key: Optional[str]) -> Optional[str]:
        """
        Return the large value according to its reference key.

        @param ref_key Reference key. This key is usually obtained through FLManager :: storeLargeValue
        @return Large value stored
        """
        if ref_key and ref_key[0:3] == "RK@":
            table_name = (
                "fllarge"
                if application.PROJECT.aq_app.singleFLLarge()
                else "fllarge_" + ref_key.split("@")[1]
            )

            conn = application.PROJECT.conn_manager.mainConn()
            cache_dir = os.path.join(application.PROJECT.tmpdir, "cache", conn.DBName(), "cacheXPM")
            if not os.path.exists(cache_dir):
                os.mkdir(cache_dir)

            file_name = os.path.join(cache_dir, "%s.xpm" % ref_key)

            if os.path.exists(file_name):
                return file_name

            if self.existsTable(table_name):
                qry = pnsqlquery.PNSqlQuery(None, "dbAux")
                qry.setSelect("contenido")
                qry.setFrom(table_name)
                qry.setWhere("refkey = '%s'" % ref_key)
                if qry.exec_() and qry.first():
                    return xpm.cache_xpm(qry.value(0), ref_key)

        return None

    def initCount(self) -> int:
        """
        Indicate the number of times FLManager :: init () has been called.
        """
        return self.init_count_



# -*- coding: utf-8 -*-
"""
Main interface with INI-Style settings.

Contains 2 global variables:

  - config: for pinebooConfig.ini
  - settings: for pinebooSettings.ini

*It's a config or a setting?*

If it's meant to be changed before Pineboo runs, it's a config.

If it's meant to be changed while running Pineboo, inside QS code, it's a setting.
"""
import json
import time
from pineboolib.core.utils import logging

from PyQt6 import QtCore  # type: ignore[import]

from typing import Dict, List, Any, Union, Tuple, Type, Optional

LOGGER = logging.get_logger(__name__)


class PinebooSettings(QtCore.QSettings):
    """Manipulate settings for the specified file."""

    CACHE_TIME_SEC = 30

    def __init__(self, name: str = "") -> None:
        """
        Build a new settings for the specified "name".

        "name" will be used for creating/opening the INI file.
        Values are saved in JSON oposed to plain text.
        """
        format_ = (
            QtCore.QSettings.Format.IniFormat
        )  # QtCore.QSettings.NativeFormat - usar solo ficheros ini.
        scope_ = QtCore.QSettings.Scope.UserScope
        self.organization = "Eneboo"
        self.application = "Pineboo%s" % name
        self.cache: Dict[str, Tuple[float, Any]] = {}
        super().__init__(format_, scope_, self.organization, self.application)

    @staticmethod
    def dump_qsize(value: "QtCore.QSize") -> Dict[str, Union[str, int]]:
        """Convert QtCore.QSize into a Dict suitable to be converted to JSON."""
        return {"__class__": "QtCore.QSize", "width": value.width(), "height": value.height()}

    def dump_value(
        self, value: Union["QtCore.QSize", str, bool, int, List[str], Dict[Any, Any]]
    ) -> str:
        """Convert Any value into JSON to be used for saving in INI."""
        if isinstance(value, QtCore.QSize):
            value = self.dump_qsize(value)
        return json.dumps(value)

    def load_value(self, value_text: str) -> Any:
        """Parse INI text values into proper Python variables."""
        value: Any = json.loads(value_text)
        if (
            isinstance(value, dict)
            and "__class__" in value.keys()
            and value["__class__"] == "QtCore.QSize"
        ):
            value = QtCore.QSize(value["width"], value["height"])

        return value

    def value(  # type: ignore [override]
        self, key: str, def_value: Any = None, type: Optional[Type] = None
    ) -> Any:
        """Get a value from INI for the specified key."""
        curtime = time.time()
        cached_value = self.cache.get(key, None)

        if cached_value is not None:  # Si tengo valor cacheado
            if curtime - cached_value[0] > self.CACHE_TIME_SEC:  # y he caducado
                del self.cache[key]
            else:
                if cached_value[1] is not None:  # Si es bueno el valor cacheado
                    return cached_value[1]

        val = self._value(key)  # repgunto por valor

        if val is not None:  # Si hay valor, lo cacheo
            self.cache[key] = (curtime, val)
        else:  # si no devuelvo default
            val = def_value
        return val

    def _value(self, key: str, default: Any = None) -> Any:
        value = super().value(key, None)
        if value is None:
            LOGGER.debug(
                "%s.value(%s) -> Default: %s %r", self.application, key, type(default), default
            )
            return default
        try:
            ret = self.load_value(value)
            LOGGER.debug("%s.value(%s) -> Loaded: %s %r", self.application, key, type(ret), ret)
            return ret
        except Exception as exc:
            # No format, just string
            LOGGER.debug("Error trying to parse json for %s: %s (%s)", key, exc, value)
            return value

    def set_value(self, key: str, value: Union["QtCore.QSize", str, bool, int, List[Any]]) -> None:
        """Set a value into INI file for specified key."""
        LOGGER.debug("%s.set_value(%s) <- %s %r", self.application, key, type(value), value)
        curtime = time.time()
        self.cache[key] = (curtime, value)
        return super().setValue(key, self.dump_value(value))

    setValue = set_value  # type: ignore [assignment]


CONFIG = PinebooSettings("Config")
SETTINGS = PinebooSettings("Settings")



"""Test_aqsbutton_group module."""

import unittest


class TestAQSButtonGroup(unittest.TestCase):
    """TestAQSButtonGroup Class."""

    def test_full(self) -> None:
        """Full test."""

        from pineboolib.fllegacy.aqsobjects import aqsbuttongroup
        from pineboolib.qsa import qsa

        button_group = aqsbuttongroup.AQSButtonGroup(None, "prueba")
        self.assertTrue(button_group.objectName(), "prueba")
        button_group2 = qsa.AQSButtonGroup(None, "prueba2")
        self.assertTrue(button_group2.objectName(), "prueba2")



# -*- coding: utf-8 -*-
"""
Module dgi_qt3ui.

Loads old Qt3 UI files and creates a Qt5 UI.
"""
from importlib import import_module

from PyQt6 import QtCore, QtGui, QtWidgets  # type: ignore[import]
from xml.etree import ElementTree as ET
from binascii import unhexlify
from pineboolib import logging
import zlib

from pineboolib.core.utils.utils_base import load2xml
from pineboolib import application
from pineboolib.application import connections


from pineboolib.q3widgets import qmainwindow, qlistview, qtoolbar, qmenu, qaction, qspinbox

from pineboolib.core import settings


from typing import Optional, Tuple, Callable, List, Dict, Any, cast, Type, Union

ICONS: Dict[str, Any] = {}
ROOT = None
LOGGER = logging.get_logger(__name__)


class Options:
    """
    Store module options.

    ***DEPRECATED***
    """

    DEBUG_LEVEL = 100


# TODO: Refactorizar este fichero como una clase. ICONS es la lista de iconos
#      para un solo formulario. Debe existir una clase LoadUI y que ICONS sea
#      una variable de ésta. Para cada nuevo formulario se debería instanciar
#      una nueva clase.


# FIXME: widget is QWidget type but Qt5-Stubs for findChild reports QObject instead of Optional[QObject]
def load_ui(form_path: str, widget: Any, parent: Optional["QtWidgets.QWidget"] = None) -> None:
    """
    Load Qt3 UI file from eneboo.

    widget: Provide a pre-created widget and this function will store UI contents on it.
    parent: Probably deprecated.
    """
    global ICONS, ROOT
    # parser = etree.XMLParser(
    #    ns_clean=True,
    #    encoding="UTF-8",
    #    remove_blank_text=True,
    # )

    tree = load2xml(form_path)

    if not tree:
        return

    ROOT = tree.getroot()
    ICONS = {}

    if parent is None:
        parent = widget

    # if application.PROJECT.DGI.localDesktop():
    widget.hide()

    for xmlimage in ROOT.findall("images//image"):
        load_icon(xmlimage)

    for xmlwidget in ROOT.findall("widget"):
        LoadWidget(xmlwidget, widget, parent)

    # print("----------------------------------")
    # for xmlwidget in ROOT.xpath("actions"):
    #     LoadWidget(xmlwidget, widget, parent)
    # print("----------------------------------")

    # Debe estar despues de LoadWidget porque queremos el valor del UI de Qt3
    formname = widget.objectName()
    LOGGER.info("form: %s", formname)

    # Cargamos actions...
    # for action in ROOT.findall("actions//action"):
    #    load_action(action, widget)

    # Cargamos menubar ...
    xmlmenubar = ROOT.find("menubar")
    # print("Cargamos menubar!")
    if xmlmenubar is not None:
        load_menu_bar(xmlmenubar, widget)

    # Cargamos toolbars ...
    # print("Cargamos toolbar!")
    for xmltoolbar in ROOT.findall("toolbars//toolbar"):
        load_tool_bar(xmltoolbar, widget)

    for xmlconnection in ROOT.findall("connections//connection"):
        sender_elem = xmlconnection.find("sender")
        signal_elem = xmlconnection.find("signal")
        receiv_elem = xmlconnection.find("receiver")
        slot_elem = xmlconnection.find("slot")

        if sender_elem is None or signal_elem is None or receiv_elem is None or slot_elem is None:
            continue

        sender_name = sender_elem.text
        signal_name = signal_elem.text
        receiv_name = receiv_elem.text
        slot_name = slot_elem.text

        if sender_name is None:
            raise ValueError("no se encuentra sender_name")
        if signal_name is None:
            raise ValueError("no se encuentra signal_name")
        if receiv_name is None:
            raise ValueError("no se encuentra receiv_name")
        if slot_name is None:
            raise ValueError("no se encuentra slot_name")

        receiver: Any = None
        if isinstance(widget, (qmainwindow.QMainWindow, QtWidgets.QMainWindow)):
            if signal_name == "activated()":
                signal_name = "triggered()"

        if sender_name == formname:
            sender = widget
        else:
            senders = widget.findChildren(QtCore.QObject, sender_name)
            for sender in senders:
                # if not application.PROJECT.DGI.localDesktop():
                #    wui = hasattr(widget, "ui_") and sender_name in widget.ui_
                #    if sender is None and wui:
                #        sender = widget.ui_[sender_name]

                sg_name = signal_name

                if signal_name.find("(") > -1:
                    sg_name = signal_name[: signal_name.find("(")]

                sl_name = slot_name
                if slot_name.find("(") > -1:
                    sl_name = slot_name[: slot_name.find("(")]

                if sender is None:
                    LOGGER.debug("Connection sender not found:%s", sender_name)

                if receiv_name == formname:
                    # fn_name = slot_name.rstrip("()")
                    fn_name = slot_name[: slot_name.find("(")]
                    LOGGER.debug(
                        "Conectando de UI a QS: (%r.%r -> %r.%r)",
                        sender_name,
                        signal_name,
                        receiv_name,
                        fn_name,
                    )

                    ifx = None
                    if not widget.inherits("QMainWindow"):
                        parent = widget.parent()
                        if parent:
                            ifx = getattr(parent, "iface", None)
                    else:
                        if sender_name in application.PROJECT.actions.keys():
                            receiver = application.PROJECT.actions[sender_name]._master_widget
                        else:
                            if receiver is None:
                                if (
                                    receiv_name == widget.__class__.__name__
                                    or receiv_name == widget.objectName()
                                ):
                                    receiver = widget

                    ifx = widget
                    # if hasattr(widget, "iface"):
                    #    ifx = widget.iface
                    if hasattr(ifx, fn_name):
                        try:
                            # getattr(sender, sg_name).connect(
                            #    getattr(ifx, fn_name))
                            connections.connect(sender, signal_name, ifx, fn_name)
                        except Exception:
                            LOGGER.debug(
                                "Error connecting: %s %s %s %s %s",
                                sender,
                                signal_name,
                                receiver,
                                slot_name,
                                getattr(ifx, fn_name),
                            )
                        continue

                if receiver is None:
                    receiver = widget.findChild(
                        QtCore.QObject,
                        receiv_name,
                        QtCore.Qt.FindChildOption.FindChildrenRecursively,
                    )

                if receiver is None:
                    from pineboolib.application.safeqsa import SafeQSA

                    receiver = SafeQSA.get_any(receiv_name)

                if receiver is None and receiv_name == "FLWidgetApplication":
                    if sender_name in application.PROJECT.actions.keys():
                        receiver = application.PROJECT.actions[sender_name]
                    else:
                        LOGGER.debug("Sender action %s not found. Connection skiped", sender_name)
                        continue

                if receiver is None:
                    LOGGER.debug("Connection receiver not found:%s", receiv_name)

                if sender is None or receiver is None:
                    continue

                if slot_name in ("openDefaultForm()", "execDefaultScript()"):
                    continue

                elif hasattr(receiver, "iface"):
                    # iface = getattr(receiver, "iface")
                    # try:
                    #    receiver.connect(
                    #        sender,
                    #        signal_name,
                    #        receiver,
                    #        "%s%s" % ("iface." if hasattr(iface, sl_name) else "", slot_name),
                    #    )
                    #     getattr(sender, sg_name).connect(getattr(iface, sl_name))
                    # except Exception:
                    #    LOGGER.exception(
                    #        "Error connecting: %s:%s %s.iface:%s", sender, signal_name, receiver, slot_name
                    #    )
                    LOGGER.warning(
                        "DEPRECATED: This type of connection must be made in the module init: %s %s %s %s",
                        sender_name,
                        signal_name,
                        receiv_name,
                        slot_name,
                    )
                    continue

                elif hasattr(receiver, sl_name):
                    try:
                        getattr(sender, sg_name).connect(getattr(receiver, sl_name))
                    except Exception:
                        LOGGER.exception(
                            "Error connecting: %s:%s %s:%s",
                            sender,
                            signal_name,
                            receiver,
                            slot_name,
                        )
                else:
                    LOGGER.error(
                        "Error connecting: %s:%s %s:%s (no candidate found)",
                        sender,
                        signal_name,
                        receiver,
                        slot_name,
                    )

    widget.show()


def load_tool_bar(xml: ET.Element, widget: QtWidgets.QMainWindow) -> None:
    """
    Load UI Toolbar from XML and store it into widget.

    widget: A pre-created widget to store the toolbar.
    """
    name_elem = xml.find("./property[@name='name']/cstring")
    label_elem = xml.find("./property[@name='label']/string")
    if name_elem is None or label_elem is None:
        raise Exception("Unable to find required name and label properties")

    name = name_elem.text or ""
    label = label_elem.text

    tool_bar = qtoolbar.QToolBar(name)
    tool_bar.setObjectName(name)
    tool_bar.label = label
    for action in xml:
        if action.tag == "action":
            name = action.get("name") or "action"
            new_action = tool_bar.addAction(name)
            if new_action:
                new_action.setObjectName(name)
                # print("**", name, new_action, tool_bar)
                load_action(action, widget, new_action)
                # print("**", new_action.objectName())
                # clone_action(new_action, widget)

            # FIXME!!, meter el icono y resto de datos!!
        elif action.tag == "separator":
            separator = tool_bar.addSeparator()
            if separator:
                separator.setObjectName("separator")
        elif action.tag == "widget":
            new_widget = WidgetResolver.get_widget_class(action.get("class") or "")(tool_bar)
            LoadWidget(action, new_widget, None, tool_bar)
            tool_bar.addWidget(new_widget)

    widget.addToolBar(tool_bar)
    widget.addToolBarBreak()


def load_menu_bar(xml: "ET.Element", widget: "QtWidgets.QWidget") -> None:
    """
    Load a menu bar into widget.

    widget: pre-created widget to store the object.
    """

    if isinstance(widget, qmainwindow.QMainWindow):
        menu_bar = widget.menuBar()
    else:
        menu_bar = QtWidgets.QMenuBar(widget)
        if menu_bar:
            widget.layout().setMenuBar(menu_bar)  # type: ignore [union-attr] # quitamos _layout()
    for item in xml:
        if item.tag == "property":
            name = item.get("name")
            if name == "name":
                cstring = item.find("cstring")
                if cstring is not None and cstring.text is not None and menu_bar:
                    menu_bar.setObjectName(cstring.text)
            elif name == "geometry":
                geo_ = item.find("rect")
                if geo_ is not None:
                    geo_x = geo_.find("x")
                    geo_y = geo_.find("y")
                    geo_width = geo_.find("width")
                    geo_height = geo_.find("height")

                    pos_x = geo_x.text if geo_x is not None else None
                    pos_y = geo_y.text if geo_y is not None else None
                    width = geo_width.text if geo_width is not None else None
                    height = geo_height.text if geo_height is not None else None
                    if pos_x is None or pos_y is None or width is None or height is None:
                        continue
                    if menu_bar:
                        menu_bar.setGeometry(int(pos_x), int(pos_y), int(width), int(height))
            elif name in ("acceptDrops", "defaultUp"):
                bool_elem = item.find("bool")
                if bool_elem is not None:
                    attr_ = getattr(menu_bar, "set%s%s" % (name[0].upper(), name[1:]))
                    attr_(bool_elem.text == "true")
            elif name == "frameShape":
                continue
            # elif name == "defaultUp":
            #    bool_elem = item.find("bool")
            #    if bool_elem is not None:
            #        menu_bar.setDefaultUp(bool_elem.text == "true")
        elif item.tag == "item" and menu_bar:
            process_item(item, menu_bar, widget)


def process_item(
    xml: "ET.Element",
    parent: Union["QtWidgets.QMenuBar", "QtWidgets.QMenu"],
    widget: "QtWidgets.QWidget",
) -> None:
    """
    Process random XML item.

    widget: pre-created widget to store the object.
    """
    name = xml.get("name") or ""
    text = xml.get("text") or ""
    # accel = xml.get("accel")

    menu_ = parent.addMenu(text)
    if menu_:
        menu_.setObjectName(name)

        for item in xml:
            if item.tag == "action":
                name_ = item.get("name") or ""
                new_action = menu_.addAction(name_)
                if new_action:
                    new_action.setObjectName(name_)

                    load_action(item, widget, new_action)
                # action.setObjectName(name_)
                # clone_action(action, widget)
            elif item.tag == "item":
                process_item(item, menu_, widget)


def clone_action(action: "QtGui.QAction", widget: "QtWidgets.QWidget") -> None:
    """
    Clone action into widget.

    widget: pre-created widget to store the object.
    used only on loadToolBar and process_item
    """
    real_action = cast(QtGui.QAction, widget.findChild(QtGui.QAction, action.objectName()))
    if real_action is not None:
        action.setText(real_action.text())
        action.setIcon(real_action.icon())
        action.setToolTip(real_action.toolTip())
        if real_action.statusTip():
            action.setStatusTip(real_action.statusTip())
        else:
            action.setStatusTip(real_action.whatsThis())
        action.setWhatsThis(real_action.whatsThis())
        action.triggered.connect(real_action.trigger)  # type: ignore [attr-defined] # noqa: F821
        action.toggled.connect(real_action.toggle)  # type: ignore [attr-defined] # noqa: F821


def load_action(
    action: "ET.Element",
    widget: "QtWidgets.QWidget",
    action_widget: Optional["QtGui.QAction"] = None,
) -> None:
    """
    Load Action into widget.

    widget: pre-created widget to store the object.
    """
    global ICONS

    new_action = action_widget if action_widget is not None else QtGui.QAction(widget)

    action_name = action.get("name")
    for root_action in ROOT.findall("actions//action"):  # type: ignore [union-attr] # noqa: F821
        for property in root_action.findall("property"):
            if property.get("name") == "name":
                cstring = property.find("cstring")
                if cstring is not None:
                    if cstring.text == action_name:
                        menu_text: Optional[str] = None
                        for property2 in root_action.findall("property"):
                            name = property2.get("name")
                            cstring = property2.find("cstring")
                            string = property2.find("string")
                            iconset = property2.find("iconset")

                            if name == "name" and cstring is not None:
                                if action_name == cstring.text:
                                    new_action.setObjectName(cstring.text or "unnamed")
                            elif name == "text" and string is not None:
                                new_action.setText(string.text or "")
                            elif name == "iconSet" and iconset is not None:
                                if iconset.text and iconset.text in ICONS.keys():
                                    new_action.setIcon(ICONS[iconset.text])
                            elif name == "toolTip" and string is not None:
                                new_action.setToolTip(string.text or "")
                            elif name == "statusTip" and string is not None:
                                new_action.setStatusTip(string.text or "")
                            elif name == "whatsThis" and string is not None:
                                new_action.setWhatsThis(string.text or "")
                            elif name == "menuText" and string is not None:
                                menu_text = string.text

                        if menu_text:
                            new_action.setText(menu_text)


class WidgetResolver:
    """
    Resolve classnames into widgets with caching.
    """

    KNOWN_WIDGETS: Dict[str, Type["QtWidgets.QWidget"]] = {}

    @classmethod
    def get_widget_class(resolver_cls, classname: str) -> Type["QtWidgets.QWidget"]:
        """Get a widget class from class name."""
        if classname in resolver_cls.KNOWN_WIDGETS:
            return resolver_cls.KNOWN_WIDGETS[classname]

        cls: Optional["QtWidgets.QWidget"] = None
        mod_name_full = "pineboolib.q3widgets.%s" % classname.lower()
        try:
            mod_ = import_module(mod_name_full)
            cls = getattr(mod_, classname, None)
        except ModuleNotFoundError:
            LOGGER.trace("resolveObject: Module not found %s", mod_name_full)
        except Exception:
            LOGGER.exception("resolveObject: Unable to load module %s", mod_name_full)

        if cls is None:
            mod_name_full = "pineboolib.fllegacy.%s" % classname.lower()
            try:
                mod_ = import_module(mod_name_full)
                cls = getattr(mod_, classname, None)
            except ModuleNotFoundError:
                LOGGER.trace("resolveObject: Module not found %s", mod_name_full)
            except Exception:
                LOGGER.exception("resolveObject: Unable to load module %s", mod_name_full)

        if cls is None:
            cls = getattr(QtWidgets, classname, None)

        if cls is None:
            raise AttributeError("Class %r not found" % classname)

        resolver_cls.KNOWN_WIDGETS[classname] = cls  # type: ignore [assignment]
        return cls  # type: ignore [return-value]


# NOTE: This function may create QAction too, which inherits from QObject, not QWidget.
def create_widget(classname: str, parent: Optional["QtWidgets.QWidget"] = None) -> "QtCore.QObject":
    """
    Create a Widget for given class name.
    """
    # FIXME: Avoid dynamic imports. Also, this is slow.
    try:
        cls = WidgetResolver.get_widget_class(classname)
        return cls(parent)
    except AttributeError:
        LOGGER.warning("WARN: Class name not found in QtWidgets:", classname)
        widgt = QtWidgets.QWidget(parent)
        widgt.setStyleSheet("* { background-color: #fa3; } ")
        return widgt


class LoadWidget:
    """Load a widget."""

    translate_properties = {
        "caption": "windowTitle",
        "name": "objectName",
        "icon": "windowIcon",
        "iconSet": "icon",
        "accel": "shortcut",
        "layoutMargin": "contentsMargins",
    }
    widget: "QtCore.QObject"
    parent: "QtWidgets.QWidget"
    orig_widget: "QtWidgets.QWidget"

    def __init__(
        self,
        xml: "ET.Element",
        widget: "QtCore.QObject",
        parent: Optional["QtCore.QObject"] = None,
        orig_widget: Optional["QtCore.QObject"] = None,
    ) -> None:
        """
        Load a random widget from given XML.
        """
        LOGGER.trace(
            "LoadWidget: xml: %s widget: %s parent: %s orig_widget: %s",
            xml,
            widget,
            parent,
            orig_widget,
        )
        if widget is None:
            raise ValueError
        if parent is None:
            parent = widget
        if orig_widget is None:
            orig_widget = widget
        self.widget = widget
        self.orig_widget = cast(QtWidgets.QWidget, orig_widget)
        self.parent = cast(QtWidgets.QWidget, parent)
        del widget
        del orig_widget
        del parent
        # if application.PROJECT.DGI.localDesktop():
        #    if not hasattr(orig_widget, "ui_"):
        #        orig_widget.ui_ = {}
        # else:
        #    orig_widget.ui_ = {}

        nwidget = None
        if self.widget == self.orig_widget:
            class_ = xml.get("class")
            if class_ is None:
                class_ = type(self.widget).__name__

            nwidget = cast(QtWidgets.QWidget, create_widget(class_, parent=self.orig_widget))
            self.parent = nwidget
        layouts_pending_process: List[Tuple[ET.Element, str]] = []
        properties = []
        unbold_fonts = []
        has_layout_defined = False

        for item in xml:
            if item.tag == "layout":
                # LOGGER.warning("Trying to replace layout. Ignoring. %s, %s", repr(c.tag), widget._layout)
                classname = item.get("class")
                if classname is None:
                    raise Exception("Expected class attr")
                lay_ = getattr(QtWidgets, classname)()
                lay_.setObjectName(item.get("name"))
                self.widget.setLayout(lay_)  # type: ignore [attr-defined] # noqa F821
                continue

            elif item.tag == "property":
                properties.append(item)
                continue

            elif item.tag in ("vbox", "hbox", "grid"):
                if (
                    has_layout_defined
                ):  # nos saltamos una nueva definición del layout ( mezclas de ui incorrectas)
                    # El primer layout que se define es el que se respeta
                    continue

                if item.tag.find("box") > -1:
                    layout_type = "Q%s%sLayout" % (item.tag[0:2].upper(), item.tag[2:])
                else:
                    layout_type = "QGridLayout"

                layout_class = WidgetResolver.get_widget_class(layout_type)
                self.widget._layout = cast(  # type: ignore [attr-defined] # noqa F821
                    QtWidgets.QLayout, layout_class()
                )

                lay_name = None
                lay_margin_v = 2
                lay_margin_h = 2
                lay_spacing = 2
                for property in item.findall("property"):
                    p_name = property.get("name")
                    number_elem = property.find("number")

                    if p_name == "name":
                        lay_name_e = property.find("cstring")
                        if lay_name_e is not None:
                            lay_name = lay_name_e.text
                    elif p_name == "margin":
                        if number_elem is not None:
                            if number_elem.text is None:
                                raise ValueError("margin no contiene valor")
                            lay_margin = int(number_elem.text)

                        if item.tag == "hbox":
                            lay_margin_h = lay_margin
                        elif item.tag == "vbox":
                            lay_margin_v = lay_margin
                        else:
                            lay_margin_h = lay_margin_v = lay_margin

                    elif p_name == "spacing":
                        if number_elem is not None:
                            if number_elem.text is None:
                                raise ValueError("spacing no contiene valor")
                            lay_spacing = int(number_elem.text)
                    elif p_name == "sizePolicy":
                        self.widget.setSizePolicy(  # type: ignore [attr-defined] # noqa F821
                            load_variant(property, self.widget)
                        )

                self.widget._layout.setSizeConstraint(  # type: ignore [attr-defined] # noqa F821
                    QtWidgets.QLayout.SizeConstraint.SetMinAndMaxSize
                )
                self.widget._layout.setObjectName(  # type: ignore [attr-defined] # noqa F821
                    lay_name or "layout"
                )
                self.widget._layout.setContentsMargins(  # type: ignore [attr-defined] # noqa F821
                    lay_margin_h, lay_margin_v, lay_margin_h, lay_margin_v
                )
                self.widget._layout.setSpacing(  # type: ignore [attr-defined] # noqa F821
                    lay_spacing
                )

                lay_type = "grid" if item.tag == "grid" else "box"
                layouts_pending_process += [(item, lay_type)]
                has_layout_defined = True
                continue

            elif item.tag == "item":
                if isinstance(self.widget, qmenu.QMenu):
                    continue
                else:
                    prop1: Dict[str, Any] = {}
                    for property_2 in item.findall("property"):
                        key, value = load_property(property_2)
                        prop1[key] = value

                    self.widget.addItem(prop1["text"])  # type: ignore [attr-defined] # noqa F821
                continue

            elif item.tag == "attribute":
                key = item.get("name")
                value = load_variant(item)
                attrs = getattr(self.widget, "_attrs", None)
                if attrs is not None:
                    attrs[key] = value
                else:
                    LOGGER.warning(
                        "qt3ui: [NOT ASSIGNED] attribute %r => %r" % (key, value),
                        self.widget.__class__,
                        repr(item.tag),
                    )
                continue
            elif item.tag == "widget":
                # Si dentro del widget hay otro significa
                # que estamos dentro de un contenedor.
                # Según el tipo de contenedor, los widgets
                # se agregan de una forma u otra.
                classname = item.get("class")
                if classname is None:
                    raise Exception("Expected class attr")
                new_widget = create_widget(classname, parent=self.parent)
                new_widget.hide()  # type: ignore [attr-defined] # noqa F821
                new_widget._attrs = {}  # type: ignore [attr-defined] # noqa F821
                LoadWidget(item, new_widget, self.parent, self.orig_widget)
                prop_name = item.find("./property[@name='name']/cstring")
                new_widget.setContentsMargins(0, 0, 0, 0)  # type: ignore [attr-defined] # noqa F821
                new_widget.show()  # type: ignore [attr-defined] # noqa F821

                if isinstance(self.widget, QtWidgets.QTabWidget):
                    title = new_widget._attrs.get(  # type: ignore [attr-defined] # noqa F821
                        "title", "UnnamedTab"
                    )
                    self.widget.addTab(cast(QtWidgets.QWidget, new_widget), title)
                elif isinstance(self.widget, (QtWidgets.QGroupBox, QtWidgets.QWidget)):
                    lay = getattr(self.widget, "layout")()
                    if not lay and not isinstance(self.widget, qtoolbar.QToolBar):
                        lay = QtWidgets.QVBoxLayout()
                        cast(QtWidgets.QWidget, self.widget).setLayout(lay)

                    if isinstance(self.widget, qtoolbar.QToolBar):
                        if isinstance(new_widget, QtGui.QAction):
                            self.widget.addAction(new_widget)
                        else:
                            self.widget.addWidget(cast(QtWidgets.QWidget, new_widget))
                    else:
                        lay.addWidget(new_widget)
                else:
                    if Options.DEBUG_LEVEL > 50:
                        LOGGER.warning(
                            "qt3ui: Unknown container widget xml tag",
                            self.widget.__class__,
                            repr(item.tag),
                        )
                unbold_fonts.append(cast(QtWidgets.QWidget, new_widget))
                continue

            elif item.tag == "action":
                action_name = item.get("name")
                if ROOT is None:
                    raise Exception("No se encuentra root")

                for xmlaction in ROOT.findall("actions//action"):
                    prop_name = xmlaction.find("./property[@name='name']/cstring")
                    if prop_name is not None and prop_name.text == action_name:
                        self.process_action(xmlaction, cast(QtWidgets.QToolBar, self.widget))
                        continue

                continue

            elif item.tag == "separator":
                cast(QtWidgets.QMenu, self.widget).addSeparator()
                continue

            elif item.tag == "column":
                for property in item.findall("property"):
                    key, value = load_property(property)
                    if key == "text":
                        cast(qlistview.QListView, self.widget).setHeaderLabel(value)
                    elif key == "clickable":
                        cast(qlistview.QListView, self.widget).setClickable(bool(value))
                    elif key == "resizable":
                        cast(qlistview.QListView, self.widget).setResizable(bool(value))

                continue

            LOGGER.info(
                "%s: Unknown widget xml tag %s %s", __name__, self.widget.__class__, repr(item.tag)
            )

        for prop in properties:
            self.process_property(prop)
        for lay, mode_ in layouts_pending_process:
            self.process_layout_box(lay, mode=mode_)
        for new_widget in unbold_fonts:
            font = new_widget.font()
            font.setBold(False)
            font.setItalic(False)
            new_widget.setFont(font)

        # if not application.PROJECT.DGI.localDesktop():
        #    if nwidget is not None and orig_widget.objectName() not in orig_widget.ui_:
        #        orig_widget.ui_[orig_widget.objectName()] = nwidget

    def process_property(self, xmlprop: ET.Element, widget_: Optional[QtCore.QObject] = None):
        """
        Process a XML property from the UI.
        """
        widget: Any
        if widget_ is None:
            widget = self.widget
        else:
            widget = widget_

        set_fn: Optional[Callable] = None
        pname = xmlprop.get("name") or ""
        pname = self.translate_properties.get(pname, pname)
        setpname = "set" + pname[0].upper() + pname[1:]
        if pname == "layoutSpacing":
            set_fn = widget._layout.setSpacing
        elif pname == "margin":
            set_fn = widget.setContentsMargins
        elif pname in ("paletteBackgroundColor", "paletteForegroundColor"):
            set_fn = widget.setStyleSheet
        elif pname == "menuText":
            if isinstance(widget, qaction.QAction):
                return
            else:
                set_fn = widget.menuText
        elif pname == "movingEnabled":
            set_fn = widget.setMovable
        elif pname == "toggleAction":
            set_fn = widget.setChecked
        elif pname == "label" and isinstance(widget, qtoolbar.QToolBar):
            return
        elif pname == "maxValue" and isinstance(widget, qspinbox.QSpinBox):
            set_fn = widget.setMaximum
        elif pname == "minValue" and isinstance(widget, qspinbox.QSpinBox):
            set_fn = widget.setMinimum
        elif pname == "lineStep" and isinstance(widget, qspinbox.QSpinBox):
            set_fn = widget.setSingleStep
        elif pname == "newLine":
            set_fn = cast(QtWidgets.QMainWindow, self.orig_widget).addToolBarBreak
        elif pname == "functionGetColor":
            set_fn = widget.setFunctionGetColor
        elif pname == "cursor":
            # Ignore "cursor" styles, this is for blinking cursor styles
            # not needed.
            return
        elif pname == "iconText":
            set_fn = widget.setWindowIconText

        else:
            set_fn = getattr(widget, setpname, None)

        if set_fn is None:
            LOGGER.warning("qt3ui: Missing property %s for %r", pname, widget.__class__)
            return

        value: Any

        if pname == "contentsMargins" or pname == "layoutSpacing":
            try:
                value = int(xmlprop.get("stdset", "0"))
                value //= 2
            except Exception:
                value = 0
            if pname == "contentsMargins":
                value = QtCore.QMargins(value, value, value, value)

        elif pname == "margin":
            try:
                value = load_variant(xmlprop)
            except Exception:
                value = 0
            value = QtCore.QMargins(value, value, value, value)

        elif pname == "paletteBackgroundColor":
            fg_color = widget.palette().color(QtGui.QPalette.ColorRole.Window).name()
            bg_color = load_variant(xmlprop).name()
            value = "color: %s; background-color: %s" % (fg_color, bg_color)

        elif pname == "paletteForegroundColor":
            bg_color = widget.palette().color(QtGui.QPalette.ColorRole.WindowText).name()
            fg_color = load_variant(xmlprop).name()

            if bg_color != fg_color:  # Evitamos todo negro
                value = "color: %s; background-color: %s" % (fg_color, bg_color)
            else:
                value = "color: %s" % fg_color

        elif pname in ["windowIcon", "icon"]:
            value1 = load_variant(xmlprop, widget)
            # FIXME: Not sure if it should return anyway
            if isinstance(value1, str):
                LOGGER.warning("Icono %s.%s no encontrado.", widget.objectName(), value1)
                return
            else:
                if value1 is None:
                    return
                value = value1
        elif pname == "buddy":
            control_name = load_variant(xmlprop, widget)
            value = None
            for child in self.orig_widget.children():
                if child.objectName() == control_name:
                    value = child
                    break
            if value is None:
                LOGGER.warning("No se encuentra %s en %s", control_name, self.orig_widget)
        else:
            value = load_variant(xmlprop, widget)
        try:
            set_fn(value)
        except Exception:
            LOGGER.exception(
                "Error processing property %s with value %s. Original XML: %s",
                pname,
                value,
                ET.tostring(xmlprop).replace(b" ", b"").replace(b"\n", b""),
            )
            # if Options.DEBUG_LEVEL > 50:
            #    print(e, repr(value))
            # if Options.DEBUG_LEVEL > 50:
            #    print(etree.ET.tostring(xmlprop))

    def process_action(self, xmlaction: ET.Element, tool_bar: QtWidgets.QToolBar):
        """
        Process a QAction.
        """
        action = cast(QtGui.QAction, create_widget("QAction"))
        for item in xmlaction:
            action_name = item.get("name")
            if action_name in self.translate_properties:
                action_name = self.translate_properties[action_name]

            self.process_property(item, action)
        tool_bar.addAction(action)
        # orig_widget.ui_[action.objectName()] = action

    def process_layout_box(self, xmllayout, widget=None, mode="box"):
        """Process layouts from UI."""
        if widget is None:
            widget = self.widget
        for item in xmllayout:
            try:
                row = int(item.get("row")) or 0
                col = int(item.get("column")) or 0
            except Exception:
                row = col = 0

            if item.tag == "property":  # Ya se han procesado previamente ...
                continue
            elif item.tag == "widget":
                new_widget = create_widget(item.get("class"), parent=widget)
                # FIXME: Should check interfaces.
                from pineboolib.q3widgets import qbuttongroup, qtoolbutton

                if isinstance(widget, qbuttongroup.QButtonGroup):
                    if isinstance(new_widget, qtoolbutton.QToolButton):
                        widget.addButton(new_widget)  # type: ignore [misc]
                        continue

                LoadWidget(item, new_widget, self.parent, self.orig_widget)
                # path = c.find("./property[@name='name']/cstring").text
                # if not application.PROJECT.DGI.localDesktop():
                #    orig_widget.ui_[path] = new_widget
                # if application.PROJECT.DGI.localDesktop():
                #    new_widget.show()
                if mode == "box":
                    try:
                        widget._layout.addWidget(new_widget)
                    except Exception:
                        LOGGER.warning(
                            "qt3ui: No se ha podido añadir %s a %s", new_widget, widget._layout
                        )

                elif mode == "grid":
                    row_span = item.get("rowspan") or 1
                    col_span = item.get("colspan") or 1
                    try:
                        widget._layout.addWidget(new_widget, row, col, int(row_span), int(col_span))
                    except Exception:
                        LOGGER.warning("qt3ui: No se ha podido añadir %s a %s", new_widget, widget)
                        LOGGER.trace("Detalle:", stack_info=True)

            elif item.tag == "spacer":
                # sH = None
                # sV = None
                hor_policy = QtWidgets.QSizePolicy.Policy.Fixed
                ver_policy = QtWidgets.QSizePolicy.Policy.Fixed
                orient_ = None
                policy_ = QtWidgets.QSizePolicy.Policy.Expanding
                row_span = item.get("rowspan") or 1
                col_span = item.get("colspan") or 1
                # policy_name = None
                spacer_name = None
                for property in item.findall("property"):
                    pname, value = load_property(property)
                    if pname == "sizeHint":
                        width = value.width()
                        height = value.height()
                    elif pname == "orientation":
                        orient_ = 2 if value == 1 else 1  # 1 Horizontal, 2 Vertical

                    elif pname == "sizeType":
                        # print("Convirtiendo %s a %s" % (p.find("enum").text, value))
                        if (
                            settings.CONFIG.value("ebcomportamiento/spacerLegacy", False)
                            or orient_ == 1
                        ):
                            policy_ = QtWidgets.QSizePolicy.Policy(value)
                        else:
                            policy_ = QtWidgets.QSizePolicy.Policy.Expanding  # Siempre Expanding

                    elif pname == "name":
                        spacer_name = value  # noqa: F841

                if orient_ == 1:
                    hor_policy = policy_
                else:
                    ver_policy = policy_

                # print(
                #    "Nuevo spacer %s (%s,%s,(%s,%s), %s, %s"
                #    % (
                #        spacer_name,
                #        "Horizontal" if orient_ == 1 else "Vertical",
                #        policy_,
                #        width,
                #        height,
                #        hor_policy,
                #        ver_policy,
                #    )
                # )
                new_spacer = QtWidgets.QSpacerItem(width, height, hor_policy, ver_policy)
                if mode == "grid":
                    widget._layout.addItem(new_spacer, row, col, int(row_span), int(col_span))
                else:
                    widget._layout.addItem(new_spacer)
                # print("Spacer %s.%s --> %s" % (spacer_name, new_spacer, widget.objectName()))
            else:
                LOGGER.warning("qt3ui: Unknown layout xml tag", repr(item.tag))

        if widget.layout() is None:
            widget.setLayout(widget._layout)
        # widget._layout.setContentsMargins(1, 1, 1, 1)
        # widget._layout.setSpacing(1)
        # widget._layout.setSizeConstraint(QtWidgets.QLayout.SetMinAndMaxSize)


def load_icon(xml: "ET.Element") -> None:
    """Load Icon from XML."""
    global ICONS

    name = xml.get("name")
    xmldata = xml.find("data")
    if name is None:
        LOGGER.warning("loadIcon: provided xml lacks attr name")
        return
    if xmldata is None:
        LOGGER.warning("loadIcon: provided xml lacks <data>")
        return
    img_format = xmldata.get("format")
    if xmldata.text is None:
        LOGGER.warning("loadIcon: text is empty")
        return

    data = unhexlify(xmldata.text.strip())
    pixmap = QtGui.QPixmap()
    if img_format == "XPM.GZ":
        data = zlib.decompress(data, 15)
        img_format = "XPM"
    pixmap.loadFromData(data, img_format)  # type: ignore [call-overload]
    icon = QtGui.QIcon(pixmap)
    ICONS[name] = icon


def load_variant(xml: ET.Element, widget: Optional[QtCore.QObject] = None) -> Any:
    """Load Variant from XML."""
    for variant in xml:
        return _load_variant(variant, widget)
    raise ValueError("No property in provided XML")


def load_property(xml: ET.Element) -> Tuple[Any, Any]:
    """Load a Qt Property from XML."""
    for variant in xml:
        return (xml.get("name"), _load_variant(variant))
    raise ValueError("No property in provided XML")


def parse_string(value: Any) -> str:
    """Convert x to string."""
    return str(value)


def parse_bool(value: str) -> bool:
    """Convert x to bool."""
    value = value.lower()
    if value[0] in ("t", "1", "on"):
        return True
    elif value[0] in ("f", "0", "off"):
        return False
    else:
        LOGGER.warning("Bool?:", repr(value))

    return False


def _load_variant(variant: ET.Element, widget: Optional[QtCore.QObject] = None) -> Any:
    """Load a variant from XM. Internal."""
    text = variant.text or ""
    text = text.strip()
    if variant.tag == "cstring":
        return text
    elif variant.tag in ["iconset", "pixmap"]:
        global ICONS
        if text in ICONS.keys():
            return ICONS[text]
        else:
            LOGGER.warning("Icon %s not found:", text)
        return
    elif variant.tag == "string":
        return parse_string(text)
    elif variant.tag == "number":
        if text.find(".") >= 0:
            return float(text)
        return int(text)
    elif variant.tag == "bool":
        return parse_bool(text)
    elif variant.tag == "rect":
        rect_ = {}
        for item in variant:
            rect_[item.tag] = int((item.text or "0").strip())
        return QtCore.QRect(rect_["x"], rect_["y"], rect_["width"], rect_["height"])

    elif variant.tag == "sizepolicy":
        policy = QtWidgets.QSizePolicy()
        for item in variant:
            ivalue_policy = int((item.text or "0").strip())
            real_policy: Any = None
            for it_ in policy.Policy:
                if it_.value == ivalue_policy:  # type: ignore [comparison-overlap]
                    real_policy = it_
                    break

            if item.tag == "hsizetype":
                policy.setHorizontalPolicy(real_policy)
            elif item.tag == "vsizetype":
                policy.setVerticalPolicy(real_policy)
            elif item.tag == "horstretch":
                policy.setHorizontalStretch(ivalue_policy)
            elif item.tag == "verstretch":
                policy.setVerticalStretch(ivalue_policy)

        return policy

    elif variant.tag == "size":
        p_sz = QtCore.QSize()
        for item in variant:
            ivalue = int((item.text or "0").strip())
            if item.tag == "width":
                p_sz.setWidth(ivalue)
            elif item.tag == "height":
                p_sz.setHeight(ivalue)
        return p_sz
    elif variant.tag == "font":
        p_font = QtGui.QFont()
        for item in variant:
            value = (item.text or "0").strip()
            bool_value: bool = False
            if item.tag not in ("family", "pointsize"):
                bool_value = parse_bool(value)
            try:
                if item.tag == "bold":
                    p_font.setBold(bool_value)
                elif item.tag == "italic":
                    p_font.setItalic(bool_value)
                elif item.tag == "family":
                    p_font.setFamily(value)
                elif item.tag == "pointsize":
                    p_font.setPointSize(int(value))
                else:
                    LOGGER.warning("unknown font style type %s", repr(item.tag))
            except Exception as exc_error:
                LOGGER.warning(exc_error)
        return p_font

    elif variant.tag == "set":
        final = 0
        text = variant.text or "0"

        if text.find("WordBreak|") > -1:
            if widget is not None and hasattr(widget, "setWordWrap"):
                widget.setWordWrap(True)  # type: ignore [attr-defined] # noqa F821
            text = text.replace("WordBreak|", "")

        for item_ in text.split("|"):
            value3 = getattr(QtCore.Qt, item_, None)
            if value3 is not None:
                final = final + int(value3)

        return QtCore.Qt.AlignmentFlag(final)

    elif variant.tag == "enum":
        libs_2: List[Any] = [
            QtCore.Qt,
            QtCore.Qt.Orientation,
            QtCore.Qt.AlignmentFlag,
            QtWidgets.QSizePolicy.Policy,
            QtWidgets.QFrame,
            QtWidgets.QSizePolicy,
            QtWidgets.QTabWidget,
            QtCore.Qt.FocusPolicy,
            QtWidgets.QFrame.Shadow,
            QtWidgets.QFrame.Shape,
            QtCore.Qt.TextFormat,
            QtWidgets.QTabWidget.TabShape,
            QtWidgets.QLineEdit.EchoMode,
        ]
        for lib in libs_2:
            value2 = getattr(lib, text, None)
            if value2 is not None:
                return value2
        if text in ["GroupBoxPanel", "LineEditPanel", "ToolBarPanel"]:
            return QtWidgets.QFrame.Shape.StyledPanel
        if text in ("Single", "SingleRow"):
            return QtWidgets.QAbstractItemView.SelectionMode.SingleSelection
        if text == "FollowStyle":
            return "QtWidgets.QTableView {selection-background-color: red;}"
        if text in ("MultiRow", "Multi"):
            return QtWidgets.QAbstractItemView.SelectionMode.MultiSelection
        if text == "Reject":
            return False

        att_found = getattr(widget, text, None)
        if att_found is not None:
            return att_found

    elif variant.tag == "color":
        qcolor = QtGui.QColor()
        red_ = 0
        green_ = 0
        blue_ = 0
        for color in variant:
            if color.text is None:
                continue
            if color.tag == "red":
                red_ = int(color.text.strip())
            elif color.tag == "green":
                green_ = int(color.text.strip())
            elif color.tag == "blue":
                blue_ = int(color.text.strip())

        qcolor.setRgb(red_, green_, blue_)
        return qcolor

    elif variant.tag == "palette":
        pal_ = QtGui.QPalette()
        for state in variant:
            print("FIXME: Procesando palette", state.tag)
            for color in state:
                red_ = 0
                green_ = 0
                blue_ = 0
                for item in color:
                    if item.text is None:
                        continue
                    if item.tag == "red":
                        red_ = int(item.text)
                    elif item.tag == "green":
                        green_ = int(item.text)
                    elif item.tag == "blue":
                        blue_ = int(item.text)

                if state.tag in ("active", "disabled", "inactive", "normal"):
                    pass
                else:
                    LOGGER.warning("Unknown palette state %s", state.tag)
                LOGGER.debug("pallete color: %s %s %s", red_, green_, blue_)

        return pal_

    elif variant.tag == "date":
        year_ = 2000
        month_ = 1
        day_ = 1
        for item_variant in variant:
            if item_variant.text is None:
                continue
            if item_variant.tag == "year":
                year_ = int(item_variant.text)
            elif item_variant.tag == "month":
                month_ = int(item_variant.text)
            elif item_variant.tag == "day":
                day_ = int(item_variant.text)

        return QtCore.QDate(year_, month_, day_)

    if Options.DEBUG_LEVEL > 50:
        LOGGER.warning("qt3ui: Unknown variant: %s --> %s ", repr(widget), ET.tostring(variant))



"""FLDatatable module."""

from pineboolib.fllegacy import fltabledb
from pineboolib import application
from pineboolib.fllegacy.tests import fixture_path

from typing import Any

import unittest
from pineboolib.loader.main import init_testing, finish_testing


class TestFLDataTable(unittest.TestCase):
    """Test FLDataTable class."""

    @classmethod
    def setUpClass(cls) -> None:
        """Ensure pineboo is initialized for testing."""
        init_testing()

    def test_basic1(self):
        """Test basic functions."""
        from pineboolib.fllegacy import systype
        import os

        qsa_sys = systype.SysType()
        path = fixture_path("principal.eneboopkg")
        self.assertTrue(os.path.exists(path))
        self.assertTrue(qsa_sys.loadModules(path, False))
        application.PROJECT.actions["flareas"].load_master_widget()

        application.PROJECT.actions["flmodules"].openDefaultForm()

        widget = application.PROJECT.actions["flmodules"]._master_widget

        if widget is None:
            self.assertTrue(widget)
            return

        form = widget.form

        if form is None:
            self.assertTrue(form)
            return

        fltable: Any = form.findChild(fltabledb.FLTableDB, "tableDBRecords")
        self.assertTrue(fltable)

        fldatatable = fltable.tableRecords()
        self.assertTrue(fldatatable)
        self.assertEqual(fldatatable.fieldName(1), "idmodulo")

        fldatatable.clearChecked()
        fldatatable.setPrimaryKeyChecked("idmodulo", True)

    @classmethod
    def tearDownClass(cls) -> None:
        """Ensure test clear all data."""
        finish_testing()



"""
PyConvert module.

Converts QS projects into Python packages that can be parsed with MyPy
"""
import sys
import os
import os.path
import codecs

import multiprocessing
from typing import List, Tuple, TypeVar, cast, Dict, Optional
from xml import etree
from pineboolib import logging
from pineboolib.core.utils import struct
from pineboolib.application.parsers import parser_qsa
from pineboolib.application.parsers.parser_qsa import postparse


LOGGER = logging.get_logger(__name__)


class Action(struct.ActionStruct):
    """Represent actions from XML."""

    modname: str = ""


ModPath = TypeVar("ModPath", bound=str)
ModName = TypeVar("ModName", bound=str)
ModList = List[Tuple[ModPath, ModName]]

CPU_COUNT: int = os.cpu_count() or 1


def _touch(path: str) -> bool:
    """Create a file if does not exist."""
    if not os.path.exists(path):
        LOGGER.info("Creating empty file %r", path)
        open(path, "a").close()
        return True
    return False


def _touch_dir(path: str) -> bool:
    """Create a folder if does not exist."""
    if not os.path.exists(path):
        LOGGER.info("Creating folder %r", path)
        os.mkdir(path)
        return True
    return False


def get_modules(from_path: str = ".") -> ModList:
    """Read folders ignoring anything suspiciuos."""
    rootdir: str = os.path.abspath(from_path)
    module_files: ModList = []
    for root, sub_folders, files in os.walk(rootdir):
        for number, subf in reversed(list(enumerate(sub_folders))):
            if subf[0] < "a" or subf[0] > "z":
                del sub_folders[number]
        root = root.replace(rootdir, "")
        if root.startswith("/"):
            root = root[1:]
        # qs_files = [fname for fname in files if fname.endswith(".qs")]
        modlist = [(root, fname.replace(".mod", "")) for fname in files if fname.endswith(".mod")]
        module_files += cast(ModList, modlist)
    return module_files


def mod_xml_parse(path: str, mod_name: str) -> Optional[Dict[str, Action]]:
    """Parse Module XML and retrieve actions."""
    try:
        tree = etree.ElementTree.parse(source=codecs.open(path, "r", encoding="iso-8859-15"))
    except Exception:
        LOGGER.exception("Error trying to parse %r", path)
        return None
    root = tree.getroot()
    actions: Dict[str, Action] = {}
    for xmlaction in root:
        action = Action(xmlaction)
        action.modname = mod_name
        if action._name in actions:
            LOGGER.warning(
                "Found duplicate action in %r for %r. Will override.", path, action._name
            )
        actions[action._name] = action
    return actions


class PythonifyItem(object):
    """Give multiprocessing something to pickle."""

    src_path: str = ""
    dst_path: str = ""
    number: int = 0
    len: int = 1
    known: Dict[str, Tuple[str, str]] = {}

    def __init__(
        self, src: str, dst: str, number: int, len: int, known: Dict[str, Tuple[str, str]]
    ):
        """Create object just from args."""
        self.src_path = src
        self.dst_path = dst
        self.number = number
        self.len = len
        self.known = known


def pythonify_item(item: PythonifyItem) -> bool:
    """Parse QS into Python. For multiprocessing.map."""
    if parser_qsa.USE_THREADS:
        LOGGER.info("(%.2f%%) Parsing QS %r", 100 * item.number / item.len, item.src_path)
    try:
        pycode = postparse.pythonify2(item.src_path, known_refs=item.known)
    except Exception:
        LOGGER.exception("El fichero %s no se ha podido convertir", item.src_path)
        return False

    file_ = codecs.open(item.dst_path, "w", encoding="UTF-8")
    file_.write(pycode)
    file_.close()

    if not os.path.exists(item.dst_path):
        LOGGER.error("The file %s was generated but doesn't exists!", item.dst_path)
        return False

    return True


def main() -> None:
    """Get options and start conversion."""
    filter_mod = sys.argv[1] if len(sys.argv) > 1 else None
    filter_file = sys.argv[2] if len(sys.argv) > 2 else None

    log_format = "%(levelname)s: %(name)s: %(message)s"
    logging.basicConfig(format=log_format, level=0)
    blib_logger = logging.get_logger("blib2to3.pgen2.driver")
    blib_logger.setLevel(logging.WARNING)
    LOGGER.info("Cpu count %d", CPU_COUNT)
    source_folder = "."
    package_name = "pymodules"
    src_path = os.path.abspath(source_folder)
    dst_path = os.path.abspath(os.path.join(src_path, package_name))
    # Step 1 - Create base package path
    _touch_dir(dst_path)
    _touch(os.path.join(dst_path, "__init__.py"))
    mypy_ini = os.path.join(src_path, "mypy.ini")
    if not os.path.exists(mypy_ini):
        with open(mypy_ini, "w", encoding="UTF-8") as file_:
            file_.write("[mypy]\n")
            file_.write("python_version = 3.7\n")
            file_.write("check_untyped_defs = True\n")

    # Step 2 - Create module folders
    module_files_in: ModList = get_modules(src_path)
    known_modules: Dict[str, Tuple[str, str]] = {}
    module_files_ok: ModList = []
    for mpath, mname in module_files_in:
        xml_name = os.path.join(mpath, "%s.xml" % mname)
        if not os.path.exists(os.path.join(src_path, xml_name)):
            LOGGER.warning("File not found %r. Ignoring module." % xml_name)
            continue
        if os.sep in mpath:
            mpath_list = mpath.split(os.sep)
            if len(mpath_list) > 2:
                LOGGER.warning("Path %r is not supported, maximum is depth 2" % mpath)
                continue
            mpath_parent = mpath_list[0]
            _touch_dir(os.path.join(dst_path, mpath_parent))
            _touch(os.path.join(dst_path, mpath_parent, "__init__.py"))

        _touch_dir(os.path.join(dst_path, mpath))
        _touch(os.path.join(dst_path, mpath, "__init__.py"))
        known_modules[mname] = (package_name + "." + mpath.replace(os.sep, "."), mname)
        module_files_ok.append((mpath, mname))

    # Step 3 - Read module XML and identify objects
    for mpath, mname in module_files_ok:
        xml_name = os.path.join(mpath, "%s.xml" % mname)
        actions = mod_xml_parse(os.path.join(src_path, xml_name), mname)
        if actions is None:
            continue
        for action in actions.values():
            if action._master_script:
                module_pubname = "form%s" % action._name
                known_modules[module_pubname] = (
                    package_name + "." + mpath.replace(os.sep, "."),
                    action._master_script,
                )
            if action._record_script:
                module_pubname = "formRecord%s" % action._name
                known_modules[module_pubname] = (
                    package_name + "." + mpath.replace(os.sep, "."),
                    action._record_script,
                )

    if filter_mod is not None:
        for alias, (path, name) in known_modules.items():
            if filter_mod not in path and filter_mod not in name:
                continue
            LOGGER.debug("from %s import %s as %s", path, name, alias)

    # Step 4 - Retrieve QS file list for conversion
    LOGGER.info("Retrieving QS File list...")
    qs_files: List[Tuple[str, str]] = []
    for mpath, mname in module_files_ok:
        if filter_mod is not None and filter_mod not in mpath and filter_mod not in mname:
            continue
        rootdir = os.path.join(src_path, mpath)
        for root, sub_folders, files in os.walk(rootdir):

            def get_fname_pair(fname: str) -> Tuple[str, str]:
                src_filename = os.path.join(root, fname)
                dst_filename = os.path.join(dst_path, mpath, fname.replace(".qs", ".py"))
                return src_filename, dst_filename

            if filter_file is not None:
                files = [fname for fname in files if filter_file in fname]
            qs_files += [get_fname_pair(fname) for fname in files if fname.endswith(".qs")]

    # Step 5 - Convert QS into Python
    LOGGER.info("Converting %d QS files...", len(qs_files))

    itemlist = [
        PythonifyItem(src=src, dst=dst, number=number, len=len(qs_files), known=known_modules)
        for number, (src, dst) in enumerate(qs_files)
    ]

    pycode_list: List[bool] = []

    if parser_qsa.USE_THREADS:
        with multiprocessing.Pool(CPU_COUNT) as cpu:
            # TODO: Add proper signatures to Python files to avoid reparsing
            pycode_list = cpu.map(pythonify_item, itemlist, chunksize=2)
    else:
        for item in itemlist:
            pycode_list.append(pythonify_item(item))

    if not all(pycode_list):
        raise Exception("Conversion failed for some files")
